/*
** Novation Automap Control Surface for Cockos Reaper
** Copyright (C) 2009 Padre P.C. (Atrabilis Sunrise / P.C.M. / info@p-c-m.org)
** License: LGPL.
**
** This software is provided 'as-is', without any express or implied
** warranty.  In no event will the authors be held liable for any damages
** arising from the use of this software.
**
** Permission is granted to anyone to use this software for any purpose,
** including commercial applications, and to alter it and redistribute it
** freely, subject to the following restrictions:
**
** 1. The origin of this software must not be misrepresented; you must not
**    claim that you wrote the original software. If you use this software
**    in a product, an acknowledgment in the product documentation would be
**    appreciated but is not required.
** 2. Altered source versions must be plainly marked as such, and must not be
**    misrepresented as being the original software.
** 3. This notice may not be removed or altered from any source distribution.
**
** Notes: the C++ interfaces used require MSVC on win32, or at least the MSVC-compatible C++ ABI. Sorry, mingw users :(
**
*/

#include "AutomapCSurf.h"
#include "AutomapActions.h"

void MidiEdit_OnCommand(int command, int flags)
{
	//HWND midiHwnd = FindWindow("REAPERmidieditorwnd", NULL);
	//if(midiHwnd)
	//	SendMessage(midiHwnd, WM_COMMAND, command, flags);

	void* midiHwnd = MIDIEditor_GetActive();
	if(midiHwnd)
		MIDIEditor_OnCommand(midiHwnd, command);
}

void KBD_OnMidiEditActionEx(int cmd, int val, int valhw, int relmode, ReaProject* proj)
{
	//HWND midiHwnd = FindWindow("REAPERmidieditorwnd", NULL);
	//if(midiHwnd)
	//	KBD_OnMainActionEx(cmd, val, valhw, relmode, midiHwnd, proj);

	HWND midiHwnd = (HWND)(MIDIEditor_GetActive());
	if(midiHwnd)
		KBD_OnMainActionEx(cmd, val, valhw, relmode, midiHwnd, proj);
}

void Main_OnExtendedCommand(const char* actionName, int flags)
{
	int commandId = plugin_register("command_id_lookup", (void*)(actionName));
	Main_OnCommand(commandId, flags);
}

double Midi2Val(int midi)
{
	return (double)(midi)/127.0;
}

int Val2Midi(double value)
{
	return (int)(127.0*value + 0.5);
}

AutomapReaperClient::AutomapReaperClient()
{
}

AutomapReaperClient::AutomapReaperClient(AutomapCSurf* parent)
: _parent(parent)
, _connection(NULL)
, _nbParams(0)
, _device(NULL)
, _instanceName("???")
{
	switch(_parent->_parameters.deviceType)
	{
		case AutomapDevice::eREMOTESL :
			_device = new AutomapRemoteSl(_parent->_parameters.hasGlobalPages);
		break;
		case AutomapDevice::eIPHONE :
			_device = new AutomapiPhone(_parent->_parameters.hasGlobalPages);
		break;
		case AutomapDevice::eNOCTURN :
			_device = new AutomapNocturn(_parent->_parameters.hasGlobalPages);
		break;
		case AutomapDevice::eREMOTESLCOMPACT :
			_device = new AutomapRemoteSlCompact(_parent->_parameters.hasGlobalPages);
		break;
		case AutomapDevice::eREMOTESLDYNAMIC :
			_device = new AutomapRemoteSlDynamic(_parent->_parameters.hasGlobalPages);
		break;
		case AutomapDevice::eREMOTESLPADRES :
			_device = new AutomapRemoteSlPadres(_parent->_parameters.hasGlobalPages);
		break;
		default:
			_device = new AutomapRemoteSl(_parent->_parameters.hasGlobalPages);
		break;
	}

	_nbParams = _device->getNbParamsPerPage();

	AutomapConnectionStatus status;
	_connection = Connect(this, &status);	// fails harmlessly if server is not present

	//! \todo Handle faulty connections: register a MessageBox() for later
	switch(status)
	{
		case CONNECTION_OK :
		break;
		case SERVER_TOO_OLD :
		break;
		case CLIENT_TOO_OLD :
		break;
		case NO_SERVER_INSTALLED :
//			MessageBox(NULL, "No Automap Server installed!", "Error!", MB_ICONEXCLAMATION | MB_OK);
		break;
		case ERROR_UNKNOWN :
		default:
		break;
	}

	_connection->Initialize();
	_connection->SetAutoFocus();
}

AutomapReaperClient::~AutomapReaperClient()
{
	Disconnect(_connection);
	_connection = NULL;

	if(_device != NULL)
	{
		delete _device;
		_device = NULL;
	}
}

void AutomapReaperClient::strcpyCrop(char* buffer, const char* source)
{
	strcpy(buffer, string(source).substr(0,AUTOMAP_NAME_LENGTH-1).c_str());
}

void AutomapReaperClient::sprintfCrop(char* buffer, const char* format, ...)
{
	//! \note BUFFER_SIZE is the absolute max !!!
	char tmp[BUFFER_SIZE];
	va_list valArguments;
	va_start(valArguments, format);
	vsprintf(tmp, format, valArguments);
	va_end(valArguments);
	strcpy(buffer, string(tmp).substr(0,AUTOMAP_NAME_LENGTH-1).c_str());
}

void AutomapReaperClient::getTrackNameCrop(MediaTrack* trackptr, char buffer[AUTOMAP_NAME_LENGTH+1], int length, int *TrackId)
{
	if(length>AUTOMAP_NAME_LENGTH)
		length = AUTOMAP_NAME_LENGTH;

	const char* title = GetTrackInfo((INT_PTR)(trackptr), NULL);
	// Always use Track view index values
	int trackIndex = CSurf_TrackToID(trackptr, false);
	if(TrackId)
		*TrackId = trackIndex;

	if(trackIndex==0)
		strcpyCrop(buffer, "Master");

	else if( title && (trackIndex>0) )
	{
		char tmp[BUFFER_SIZE];
		sprintf(tmp, "T%d ", trackIndex);
		string titleshort = tmp;
		titleshort += title;
		strcpyCrop(buffer, titleshort.substr(0,length-1).c_str());
	}

	else
		strcpyCrop(buffer, "???");
}

AutomapDevice* AutomapReaperClient::getDevice()
{
	return _device;
}

void AutomapReaperClient::GetName(char *name) const
{
	AutomapReaperClient::strcpyCrop(name, "Reaper");
}

void AutomapReaperClient::GetMfr(char *mfr) const
{
	AutomapReaperClient::strcpyCrop(mfr, "Cockos");
}

void AutomapReaperClient::GetType(char *type) const
{
	AutomapReaperClient::strcpyCrop(type, "Mixer");
}

void AutomapReaperClient::GetSaveName(char *name) const
{
	AutomapReaperClient::strcpyCrop(name, "Reaper");
}

Automap::GUID AutomapReaperClient::GetGUID() const
{
	return Automap::GUID("{666988CB-0C2B-4BCD-8B24-0FFEE57870B0}");
}

void AutomapReaperClient::GetInstanceName(char *name) const
{
	AutomapReaperClient::strcpyCrop(name, _instanceName.c_str());
}

void AutomapReaperClient::SaveInstanceName(const char *name)
{
}

int AutomapReaperClient::GetNumParams() const
{
	return _nbParams;
}

Automap::Connection* AutomapReaperClient::GetConnection()
{
	return _connection;
}

bool AutomapReaperClient::GetParamInfo(int idx, Automap::ParamInfo& data)
{
	AutomapReaperClient::strcpyCrop(data.name, "");

	AutomapController* controller = _device->getController(idx);
	if(controller == NULL)
		return false;

	data.maxInteger = controller->maxInteger;
	data.minInteger = controller->minInteger;
	data.stepInteger = controller->stepInteger;

	int trackIndex = controller->getTrackIndex();
	if(trackIndex<0)
		trackIndex = _parent->getSelTrack();

	MediaTrack* tr = CSurf_TrackFromID(trackIndex, _parent->getMcpMode());
	char title[AUTOMAP_NAME_LENGTH+1] = "???";

	int realTrackIndex = trackIndex;
	if(tr)
		getTrackNameCrop(tr, title, AUTOMAP_NAME_LENGTH-10, &realTrackIndex);

	switch(controller->getType())
	{
		case eCTRL_CURSEL_VOL :
		case eCTRL_CURSEL_PAN :
		case eCTRL_CURSEL_MUTE :
		case eCTRL_CURSEL_SOLO :
		case eCTRL_CURSEL_ARM :
		case eCTRL_CURSEL_PHASE :
		case eCTRL_CURSEL_FXBYPASS :
		case eCTRL_CURSEL_FXSELECT_WHEEL :
		case eCTRL_CURSEL_FXPARAMSELECT_WHEEL :
		case eCTRL_CURSEL_TRACKSENDSELECT_WHEEL :
		case eCTRL_CURSEL_TRACKHARDSENDSELECT_WHEEL :
		case eCTRL_SUBMENUSELECT_WHEEL :
		case eCTRL_FXPARAM :
		case eCTRL_CURSEL_NAME :
		case eCTRL_CURSEL_VIEW_ENVELOPES :
		case eCTRL_CURSEL_VIEW_FX :
		case eCTRL_CURSEL_VIEW_IO :
		case eCTRL_CURSEL_VIEW_GROUPING :
		case eCTRL_CURSEL_AUTOMATION_MODE :
			if(realTrackIndex<0)
				return false;
		break;
		default:
		break;
	}

	switch(controller->getType())
	{
		case eCTRL_VOL		: AutomapReaperClient::sprintfCrop(data.name, "%s Vol", title);		break;
		case eCTRL_PAN		: AutomapReaperClient::sprintfCrop(data.name, "%s Pan", title);		break;
		case eCTRL_MUTE		: AutomapReaperClient::sprintfCrop(data.name, "%s Mute", title);	break;
		case eCTRL_PHASE	: AutomapReaperClient::sprintfCrop(data.name, "%s Phase", title);	break;
		case eCTRL_SOLO		: AutomapReaperClient::sprintfCrop(data.name, "%s Solo", title);	break;
		case eCTRL_ARM		: AutomapReaperClient::sprintfCrop(data.name, "%s Arm", title);		break;

		case eCTRL_CURSEL_VOL				: AutomapReaperClient::sprintfCrop(data.name, "T%d Vol", realTrackIndex);		break;
		case eCTRL_CURSEL_PAN				: AutomapReaperClient::sprintfCrop(data.name, "T%d Pan", realTrackIndex);		break;
		case eCTRL_CURSEL_MUTE				: AutomapReaperClient::sprintfCrop(data.name, "T%d Mute", realTrackIndex);		break;
		case eCTRL_CURSEL_PHASE				: AutomapReaperClient::sprintfCrop(data.name, "T%d Phase", realTrackIndex);		break;
		case eCTRL_CURSEL_SOLO				: AutomapReaperClient::sprintfCrop(data.name, "T%d Solo", realTrackIndex);		break;
		case eCTRL_CURSEL_ARM				: AutomapReaperClient::sprintfCrop(data.name, "T%d Arm", realTrackIndex);		break;
		case eCTRL_CURSEL_FXBYPASS			: AutomapReaperClient::sprintfCrop(data.name, "T%d FX", realTrackIndex);		break;
		case eCTRL_CURSEL_NAME				: AutomapReaperClient::sprintfCrop(data.name, "t:Name:");						break;
		case eCTRL_CURSEL_VIEW_ENVELOPES	: AutomapReaperClient::strcpyCrop(data.name, "t:Env");							break;
		case eCTRL_CURSEL_VIEW_FX			: AutomapReaperClient::strcpyCrop(data.name, "t:FX");							break;
		case eCTRL_CURSEL_VIEW_IO			: AutomapReaperClient::strcpyCrop(data.name, "t:I/O");							break;
		case eCTRL_CURSEL_VIEW_GROUPING		: AutomapReaperClient::strcpyCrop(data.name, "t:Groups");						break;
		case eCTRL_CURSEL_AUTOMATION_MODE	: AutomapReaperClient::strcpyCrop(data.name, "e:Mode");							break;

		case eCTRL_TRACKSELECT			: AutomapReaperClient::sprintfCrop(data.name, "T%d Select", realTrackIndex);	break;
		case eCTRL_TRACKSELECT_WHEEL	: AutomapReaperClient::strcpyCrop(data.name, "Track");							break;
		case eCTRL_MCPMODE				: AutomapReaperClient::strcpyCrop(data.name, "T/MCP");							break;

		case eCTRL_HZOOM_WHEEL : case eCTRL_HZOOM_POT	: AutomapReaperClient::strcpyCrop(data.name, "HZoom");				break;
		case eCTRL_VZOOM_WHEEL : case eCTRL_VZOOM_POT	: AutomapReaperClient::strcpyCrop(data.name, "VZoom");				break;

		case eCTRL_SCRUB_JOG_WHEEL : case eCTRL_SCRUB_JOG_POT :
			AutomapReaperClient::strcpyCrop(data.name, "Scrub");
		break;

		case eCTRL_TCPFXPARAM : case eCTRL_TCPFXPARAM_ENC :
			AutomapReaperClient::sprintfCrop(data.name, "FX Param %02d", _parent->getSelTcpFxParam()+controller->getSlot(idx));
		break;

		case eCTRL_PLAY			: AutomapReaperClient::strcpyCrop(data.name, "Play");		break;
		case eCTRL_STOP			: AutomapReaperClient::strcpyCrop(data.name, "Stop");		break;
		case eCTRL_RECORD		: AutomapReaperClient::strcpyCrop(data.name, "Record");		break;
		case eCTRL_FFW			: AutomapReaperClient::strcpyCrop(data.name, ">>");			break;
		case eCTRL_RWD			: AutomapReaperClient::strcpyCrop(data.name, "<<");			break;
		case eCTRL_GOSTART		: AutomapReaperClient::strcpyCrop(data.name, "|<<");		break;
		case eCTRL_GOEND		: AutomapReaperClient::strcpyCrop(data.name, ">>|");		break;
		case eCTRL_LOOP			: AutomapReaperClient::strcpyCrop(data.name, "Loop");		break;

		case eCTRL_ASSIGNABLEFN	:
		{
			const char* alias = AutomapCSurf::getFnButAlias(controller->getSlot(idx));
			if(alias)
				AutomapReaperClient::strcpyCrop(data.name, alias);
			else
				AutomapReaperClient::sprintfCrop(data.name, "Fn%02d", controller->getSlot(idx)+1);
		}
		break;

		case eCTRL_ASSIGNABLEFNPOT	:
		{
			const char* alias = AutomapCSurf::getFnPotAlias(controller->getSlot(idx));
			if(alias)
				AutomapReaperClient::strcpyCrop(data.name, alias);
			else
				AutomapReaperClient::sprintfCrop(data.name, "FnPot%02d", controller->getSlot(idx)+1);
		}
		break;

		case eCTRL_ASSIGNABLEFNENC	:
		{
			const char* alias = AutomapCSurf::getFnEncAlias(controller->getSlot(idx));
			if(alias)
				AutomapReaperClient::strcpyCrop(data.name, alias);
			else
				AutomapReaperClient::sprintfCrop(data.name, "FnEnc%02d", controller->getSlot(idx)+1);
		}
		break;

		case eCTRL_MARKER_PREV		: AutomapReaperClient::strcpyCrop(data.name, "<<-Mkr");			break;
		case eCTRL_MARKER_NEXT		: AutomapReaperClient::strcpyCrop(data.name, "Mkr->>");			break;
		case eCTRL_LOOP_SETSTART	: AutomapReaperClient::strcpyCrop(data.name, "SetLpStrt");		break;
		case eCTRL_LOOP_SETEND		: AutomapReaperClient::strcpyCrop(data.name, "SetLpEnd");		break;
		case eCTRL_INSERT_MARKERRGN	: AutomapReaperClient::strcpyCrop(data.name, "Add Rgn");		break;
		case eCTRL_INSERT_MARKER	: AutomapReaperClient::strcpyCrop(data.name, "Add Mrkr");		break;
		case eCTRL_DELETE_MARKER	: AutomapReaperClient::strcpyCrop(data.name, "Del Mrkr");		break;
		case eCTRL_CLEARRECARM		: AutomapReaperClient::strcpyCrop(data.name, "ClrRecArm");		break;

		case eCTRL_FFWRWD_WHEEL								: AutomapReaperClient::strcpyCrop(data.name, "Ffw/Rwd");		break;
		case eCTRL_PREVNEXTMARKER_WHEEL						: AutomapReaperClient::strcpyCrop(data.name, "Markers");		break;
		case eCTRL_CURSEL_FXSELECT_WHEEL					: AutomapReaperClient::strcpyCrop(data.name, "FX");				break;
		case eCTRL_CURSEL_FXPARAMSELECT_WHEEL				: AutomapReaperClient::strcpyCrop(data.name, "FX Param");		break;
		case eCTRL_CURSEL_TCPFXPARAMSELECT_WHEEL			: AutomapReaperClient::strcpyCrop(data.name, "FX Param");		break;
		case eCTRL_CURSEL_TRACKSENDSELECT_WHEEL				: AutomapReaperClient::strcpyCrop(data.name, "Send");			break;
		case eCTRL_CURSEL_TRACKHARDSENDSELECT_WHEEL			: AutomapReaperClient::strcpyCrop(data.name, "HSend");			break;
		case eCTRL_VSCROLL_WHEEL : case eCTRL_VSCROLL_POT	:	AutomapReaperClient::strcpyCrop(data.name, "VScroll");		break;
		case eCTRL_HSCROLL_WHEEL : case eCTRL_HSCROLL_POT	:	AutomapReaperClient::strcpyCrop(data.name, "HScroll");		break;

		case eCTRL_SUBMENUSELECT_WHEEL				: AutomapReaperClient::strcpyCrop(data.name, "Control");		break;

		case eCTRL_FXPARAM :
		{
			int fxParamOffset = controller->getType() - eCTRL_FXPARAM;
			int fxParam = _parent->getSelFxParam()+fxParamOffset;
			fxParam += controller->getSlot(idx);

			if(tr)
			{
				char buffer[BUFFER_SIZE];
				bool res = TrackFX_GetParamName(tr, _parent->getSelFx(), fxParam, buffer, AUTOMAP_NAME_LENGTH);
				if(res)
					AutomapReaperClient::strcpyCrop(data.name, buffer);
				else
					return false;
			}
		}
		break;

		case eCTRL_TRACKSEND_VOL :
		case eCTRL_TRACKSEND_PAN :
		case eCTRL_TRACKSEND_MUTE :
		case eCTRL_TRACKSEND_MODE :
		case eCTRL_TRACKSEND_MUTE_POT :
		case eCTRL_TRACKSEND_MODE_POT :
		case eCTRL_TRACKHARDSEND_VOL :
		case eCTRL_TRACKHARDSEND_PAN :
		case eCTRL_TRACKHARDSEND_MUTE :
		case eCTRL_TRACKHARDSEND_MODE :
		case eCTRL_TRACKHARDSEND_MUTE_POT :
		case eCTRL_TRACKHARDSEND_MODE_POT :
			if(!getTrackSendParamName(tr, controller->getType(), controller->getSlot(idx), data.name))
				return false;
		break;

		case eMIDICTRL_VZOOM_WHEEL				: AutomapReaperClient::strcpyCrop(data.name, "m:VZoom");		break;
		case eMIDICTRL_HZOOM_WHEEL				: AutomapReaperClient::strcpyCrop(data.name, "m:HZoom");		break;
		case eMIDICTRL_VSCROLL_WHEEL			: AutomapReaperClient::strcpyCrop(data.name, "m:VScrol");		break;
		case eMIDICTRL_HSCROLL_WHEEL			: AutomapReaperClient::strcpyCrop(data.name, "m:HScrol");		break;
		case eMIDICTRL_VMOVE_WHEEL				: AutomapReaperClient::strcpyCrop(data.name, "m:VMove");		break;
		case eMIDICTRL_HMOVE_WHEEL				: AutomapReaperClient::strcpyCrop(data.name, "m:HMove");		break;
		case eMIDICTRL_GRID_WHEEL				: AutomapReaperClient::strcpyCrop(data.name, "m:Grid");			break;
		case eMIDICTRL_TOGGLE_VIEWMODE			: AutomapReaperClient::strcpyCrop(data.name, "m:View");			break;
		case eMIDICTRL_TOGGLE_SNAP				: AutomapReaperClient::strcpyCrop(data.name, "m:Snap");			break;
		case eMIDICTRL_TOGGLE_GRID				: AutomapReaperClient::strcpyCrop(data.name, "m:Grid");			break;
		case eMIDICTRL_MOVE_UP					: AutomapReaperClient::strcpyCrop(data.name, "m:^");			break;
		case eMIDICTRL_MOVE_DOWN				: AutomapReaperClient::strcpyCrop(data.name, "m:v");			break;
		case eMIDICTRL_MOVE_RIGHT				: AutomapReaperClient::strcpyCrop(data.name, "m:>>");			break;
		case eMIDICTRL_MOVE_LEFT				: AutomapReaperClient::strcpyCrop(data.name, "m:<<");			break;
		case eMIDICTRL_INSERTNOTE				: AutomapReaperClient::strcpyCrop(data.name, "m:Insert");		break;
		case eMIDICTRL_NOTEPROPERTIES			: AutomapReaperClient::strcpyCrop(data.name, "m:Prop");			break;
		case eMIDICTRL_NEXTSECONDARY			: AutomapReaperClient::strcpyCrop(data.name, "M:Next");			break;
		case eMIDICTRL_VELOCITY_INC_FINE		: AutomapReaperClient::strcpyCrop(data.name, "m:Vel:+1");		break;
		case eMIDICTRL_VELOCITY_INC_COARSE		: AutomapReaperClient::strcpyCrop(data.name, "m:Vel:+10");		break;
		case eMIDICTRL_VELOCITY_DEC_FINE		: AutomapReaperClient::strcpyCrop(data.name, "m:Vel:-1");		break;
		case eMIDICTRL_VELOCITY_DEC_COARSE		: AutomapReaperClient::strcpyCrop(data.name, "m:Vel:-10");		break;
		case eMIDICTRL_VELOCITY_FINE_WHEEL		: AutomapReaperClient::strcpyCrop(data.name, "m:VelFn");		break;
		case eMIDICTRL_VELOCITY_COARSE_WHEEL	: AutomapReaperClient::strcpyCrop(data.name, "m:VelCrs");		break;
		case eMIDICTRL_TRANSPOSE_INC1			: AutomapReaperClient::strcpyCrop(data.name, "m:+1");			break;
		case eMIDICTRL_TRANSPOSE_DEC1			: AutomapReaperClient::strcpyCrop(data.name, "m:-1");			break;
		case eMIDICTRL_TRANSPOSE_INC12			: AutomapReaperClient::strcpyCrop(data.name, "m:+12");			break;
		case eMIDICTRL_TRANSPOSE_DEC12			: AutomapReaperClient::strcpyCrop(data.name, "m:-12");			break;
		case eMIDICTRL_HUMANIZE					: AutomapReaperClient::strcpyCrop(data.name, "m:Human");		break;
		case eMIDICTRL_FIT_CONTENT				: AutomapReaperClient::strcpyCrop(data.name, "m:FitCnt");		break;
		case eMIDICTRL_FIT_LOOP					: AutomapReaperClient::strcpyCrop(data.name, "m:FitLp");		break;
		case eMIDICTRL_DELETENOTE				: AutomapReaperClient::strcpyCrop(data.name, "m:Delete");		break;
		case eMIDICTRL_NOTE_MOVELEFT			: AutomapReaperClient::strcpyCrop(data.name, "m:<-");			break;
		case eMIDICTRL_NOTE_MOVERIGHT			: AutomapReaperClient::strcpyCrop(data.name, "m:->");			break;
		case eMIDICTRL_NOTE_SELECTNEXT			: AutomapReaperClient::strcpyCrop(data.name, "m:->>");			break;
		case eMIDICTRL_NOTE_SELECTPREV			: AutomapReaperClient::strcpyCrop(data.name, "m:<<-");			break;
		case eMIDICTRL_NOTE_VALUE_SELNEXT		: AutomapReaperClient::strcpyCrop(data.name, "m:Val->>");		break;
		case eMIDICTRL_NOTE_VALUE_SELPREV		: AutomapReaperClient::strcpyCrop(data.name, "m:<<-Val");		break;
		case eMIDICTRL_NOTE_LENGTH_WHEEL		: AutomapReaperClient::strcpyCrop(data.name, "m:Length");		break;
		case eMAINMIDICTRL_INSERTITEM			: AutomapReaperClient::strcpyCrop(data.name, "M:Ins");			break;
		case eMAINMIDICTRL_OPENITEMS			: AutomapReaperClient::strcpyCrop(data.name, "M:Open");			break;
		case eMIDICTRL_CCLANE_POT				: AutomapReaperClient::strcpyCrop(data.name, "m:CC Lane");		break;

		case eCTRL_GRID_WHEEL				: AutomapReaperClient::strcpyCrop(data.name, "Grid");			break;
		case eCTRL_GROUPING_OVERRIDE		: AutomapReaperClient::strcpyCrop(data.name, "Group");			break;
		case eCTRL_GRID_LINES				: AutomapReaperClient::strcpyCrop(data.name, "Grid");			break;
		case eCTRL_METRONOME				: AutomapReaperClient::strcpyCrop(data.name, "Metronome");		break;
		case eCTRL_SNAP						: AutomapReaperClient::strcpyCrop(data.name, "Snap");			break;
		case eCTRL_MIXER_VISIBLE			: AutomapReaperClient::strcpyCrop(data.name, "Mixer");			break;
		case eCTRL_TRACK_GROUPING_MATRIX	: AutomapReaperClient::strcpyCrop(data.name, "GrpMat");			break;
		case eCTRL_PERFORMANCE_METER		: AutomapReaperClient::strcpyCrop(data.name, "Perf");			break;
		case eCTRL_ROUTING_MATRIX			: AutomapReaperClient::strcpyCrop(data.name, "RtgMat");			break;
		case eCTRL_NAVIGATOR_WINDOW			: AutomapReaperClient::strcpyCrop(data.name, "Navig");			break;
		case eCTRL_METRONOME_VOLUME			: AutomapReaperClient::strcpyCrop(data.name, "Metro Vol");		break;
		case eCTRL_SOLO_INFRONT_ADJUST_POT	: AutomapReaperClient::strcpyCrop(data.name, "SoloFnt");		break;
	
		case eCTRL_TRACK_FREEPOS				: AutomapReaperClient::strcpyCrop(data.name, "t:FrPos");		break;
		case eCTRL_ITEM_MOVE_DOWN				: AutomapReaperClient::strcpyCrop(data.name, "i:v");			break;
		case eCTRL_ITEM_MOVE_UP					: AutomapReaperClient::strcpyCrop(data.name, "i:^");			break;
		case eCTRL_ITEM_MOVE_RIGHT				: AutomapReaperClient::strcpyCrop(data.name, "i:->");			break;
		case eCTRL_ITEM_MOVE_LEFT				: AutomapReaperClient::strcpyCrop(data.name, "i:<-");			break;
		case eCTRL_ITEM_GROUP					: AutomapReaperClient::strcpyCrop(data.name, "i:Group");		break;
		case eCTRL_ITEM_SELECTALL_INGROUP		: AutomapReaperClient::strcpyCrop(data.name, "i:SelGrp");		break;
		case eCTRL_ITEM_REMOVE_FROMGROUP		: AutomapReaperClient::strcpyCrop(data.name, "i:UnGrp");		break;
		case eCTRL_ITEM_SELECTALL_INTIMESEL		: AutomapReaperClient::strcpyCrop(data.name, "i:SelTim");		break;
		case eCTRL_ITEM_SELECTALL_INSELTRACK	: AutomapReaperClient::strcpyCrop(data.name, "i:SelTrk");		break;
		case eCTRL_ITEM_UNSELECTALL				: AutomapReaperClient::strcpyCrop(data.name, "i:Unsel");		break;
		case eCTRL_ITEM_PITCH_ONECENT_DOWN		: AutomapReaperClient::strcpyCrop(data.name, "i:Pitch-");		break;
		case eCTRL_ITEM_PITCH_OCTAVE_DOWN		: AutomapReaperClient::strcpyCrop(data.name, "i:Pitch-12");		break;
		case eCTRL_ITEM_PITCH_SEMITONE_DOWN		: AutomapReaperClient::strcpyCrop(data.name, "i:Pitch-1");		break;
		case eCTRL_ITEM_PITCH_ONECENT_UP		: AutomapReaperClient::strcpyCrop(data.name, "i:Pitch+");		break;
		case eCTRL_ITEM_PITCH_OCTAVE_UP			: AutomapReaperClient::strcpyCrop(data.name, "i:Pitch+12");		break;
		case eCTRL_ITEM_PITCH_SEMITONE_UP		: AutomapReaperClient::strcpyCrop(data.name, "i:Pitch+1");		break;
		case eCTRL_ITEM_SELECT_NEXTRACK			: AutomapReaperClient::strcpyCrop(data.name, "i:SelNxtTr");		break;
		case eCTRL_ITEM_SELECT_PREVTRACK		: AutomapReaperClient::strcpyCrop(data.name, "i:SelPrvTr");		break;
		case eCTRL_ITEM_SELECT_NEXTITEM			: AutomapReaperClient::strcpyCrop(data.name, "i:->>");			break;
		case eCTRL_ITEM_SELECT_PREVITEM			: AutomapReaperClient::strcpyCrop(data.name, "i:<<-");			break;
		case eCTRL_TAKE_ACTIVE_CROP				: AutomapReaperClient::strcpyCrop(data.name, "tk:Crop");		break;
		case eCTRL_TAKE_ACTIVE_DELETE			: AutomapReaperClient::strcpyCrop(data.name, "tk:Del");			break;
		case eCTRL_TAKE_ACTIVE_NEXT				: AutomapReaperClient::strcpyCrop(data.name, "tk:->>");			break;
		case eCTRL_TAKE_ACTIVE_PREVIOUS			: AutomapReaperClient::strcpyCrop(data.name, "tk:<<-");			break;
		case eCTRL_TAKE_REVERSE					: AutomapReaperClient::strcpyCrop(data.name, "tk:Rev");			break;

		case eCTRL_ITEM_PITCH_ONECENT_WHEEL		: AutomapReaperClient::strcpyCrop(data.name, "i:PitCnt");		break;
		case eCTRL_ITEM_PITCH_OCTAVE_WHEEL		: AutomapReaperClient::strcpyCrop(data.name, "i:PitOct");		break;
		case eCTRL_ITEM_PITCH_SEMITONE_WHEEL	: AutomapReaperClient::strcpyCrop(data.name, "i:Pitch");		break;

		case eCTRL_ITEM_MUTE			: AutomapReaperClient::strcpyCrop(data.name, "i:Mute");		break;
		case eCTRL_ITEM_LOOP_SOURCE		: AutomapReaperClient::strcpyCrop(data.name, "i:LpSrc");	break;
		case eCTRL_ITEM_LOOP_SECTION	: AutomapReaperClient::strcpyCrop(data.name, "i:LpSec");	break;
		case eCTRL_ITEM_PROPERTIES		: AutomapReaperClient::strcpyCrop(data.name, "i:Prop");		break;

		case eCTRL_TAKESELECT_WHEEL		: AutomapReaperClient::strcpyCrop(data.name, "Take");		break;
		case eCTRL_ITEMSELECT_WHEEL		: AutomapReaperClient::strcpyCrop(data.name, "Item");		break;

		case eCTRL_SHOW_ALL_ACTIVE_ENVELOPES	: AutomapReaperClient::strcpyCrop(data.name, "e:Show");		break;
		case eCTRL_ENVELOPE_BYPASS				: AutomapReaperClient::strcpyCrop(data.name, "e:Bypass");	break;
		case eCTRL_ENVELOPE_RECARM				: AutomapReaperClient::strcpyCrop(data.name, "e:Arm");		break;
		case eCTRL_ENVELOPE_INSERT_POINT		: AutomapReaperClient::strcpyCrop(data.name, "e:Insert");	break;

		case eCTRL_ENVELOPE_DEFAULT_SHAPE	: AutomapReaperClient::strcpyCrop(data.name, "e:DShape");		break;
		case eCTRL_ENVELOPE_SELECTED_SHAPE	: AutomapReaperClient::strcpyCrop(data.name, "e:Shape");		break;

		case eCTRL_ENVELOPE_VISIBLE_VOLUME		: AutomapReaperClient::strcpyCrop(data.name, "e:Volume");		break;
		case eCTRL_ENVELOPE_VISIBLE_PAN			: AutomapReaperClient::strcpyCrop(data.name, "e:Pan");			break;
		case eCTRL_ENVELOPE_VISIBLE_PREFXVOL	: AutomapReaperClient::strcpyCrop(data.name, "e:p-Vol");		break;
		case eCTRL_ENVELOPE_VISIBLE_PREFXPAN	: AutomapReaperClient::strcpyCrop(data.name, "e:p-Pan");		break;

		case eCTRL_ENVELOPE_CLEAR			: AutomapReaperClient::strcpyCrop(data.name, "e:Clear");		break;
		case eCTRL_ENVELOPE_DELETE_POINTS	: AutomapReaperClient::strcpyCrop(data.name, "e:Delete");		break;
		case eCTRL_ENVELOPE_TOGGLE_DISPLAY	: AutomapReaperClient::strcpyCrop(data.name, "e:Hide");			break;
		case eCTRL_ENVELOPE_ALLVISIBLE_LANE	: AutomapReaperClient::strcpyCrop(data.name, "e:Lane");			break;

		case eCTRL_AUTOMATION_ACTIVE_FADER_POT		: AutomapReaperClient::strcpyCrop(data.name, "e:Fader");		break;
		case eCTRL_AUTOMATION_ACTIVE_FADER_WHEEL	: AutomapReaperClient::strcpyCrop(data.name, "e:Fader");		break;

		case eCTRL_UNDO_HISTORY	: AutomapReaperClient::strcpyCrop(data.name, "Undos");		break;

		case eCTRL_SWS_TRACKLIST		: AutomapReaperClient::strcpyCrop(data.name, "Tracks");			break;
		case eCTRL_SWS_MARKERLIST		: AutomapReaperClient::strcpyCrop(data.name, "Markers");		break;
		case eCTRL_SWS_SNAPSHOTLIST		: AutomapReaperClient::strcpyCrop(data.name, "SnpShots");		break;

		case eCTRL_SWS_ZOOM_TO_TRACKTIMESELECT	: AutomapReaperClient::strcpyCrop(data.name, "ZoomSel");		break;
		case eCTRL_SWS_NEW_SNAPSHOT				: AutomapReaperClient::strcpyCrop(data.name, "NewSnap");		break;
		case eCTRL_SWS_SAVE_SNAPSHOT			: AutomapReaperClient::strcpyCrop(data.name, "SavSnap");		break;

		case eCTRL_MOVE_CURSOR_GRID_WHEEL	: AutomapReaperClient::strcpyCrop(data.name, "Cursor");			break;

		case eCTRL_SHOW_ACTIONLIST			: AutomapReaperClient::strcpyCrop(data.name, "Actions");		break;

		case eCTRL_ITEM_GLUED_SELECTED		: AutomapReaperClient::strcpyCrop(data.name, "i:Glue");			break;

		case eCTRL_SPEED_DIAL				: AutomapReaperClient::strcpyCrop(data.name, "Speed Dial");		break;

		default :
			AutomapReaperClient::strcpyCrop(data.name, "");
			return false;
		break;
	}

	return true;
}

float AutomapReaperClient::GetParamValue(int idx, char valueTextOut[AUTOMAP_NAME_LENGTH+1])
{
	double value = 0.0;
	if(valueTextOut)
		AutomapReaperClient::strcpyCrop(valueTextOut, " ");

	AutomapController* controller = _device->getController(idx);
	if(controller == NULL)
	{
		if(valueTextOut)
			AutomapReaperClient::strcpyCrop(valueTextOut, " ");
		return 0.0;
	}

	int trackIndex = controller->getTrackIndex();
	if(trackIndex<0)
		trackIndex = _parent->getSelTrack();
	MediaTrack* tr = CSurf_TrackFromID(trackIndex, _parent->getMcpMode());

	double volume = 0.0;
	double pan = 0.0;
	if(tr)
		GetTrackUIVolPan(tr, &volume, &pan);

	int flags = 0;
	if(tr)
		GetTrackInfo((INT_PTR)(tr), &flags);
	bool soloed = ((flags&16) != 0);
	bool muted = ((flags&8) != 0);
	bool recarmed = ((flags&64) != 0);
	bool fx = ((flags&4) != 0);

	flags = GetPlayState();
	bool playing = ((flags&1) != 0);
	bool pause = ((flags&2) != 0);
	bool recording = ((flags&4) != 0);

	switch(controller->getType())
	{
		case eCTRL_VOL :
		case eCTRL_CURSEL_VOL :
			if(tr)
				value = vol2Value(volume, valueTextOut);
		break;

		case eCTRL_PAN :
		case eCTRL_CURSEL_PAN :
			if(tr)
				value = pan2Value(pan, valueTextOut);
		break;

		case eCTRL_MUTE :
		case eCTRL_CURSEL_MUTE :
			if(tr)
				value = boolStatus2Value(muted, valueTextOut);
		break;

		case eCTRL_PHASE :
		case eCTRL_CURSEL_PHASE :
			if(tr)
			{
				bool phased = false;
				void* tmp = GetSetMediaTrackInfo(tr, "B_PHASE", NULL);
				if(tmp)
					phased = *(bool*)(tmp);
				value = boolStatus2Value(phased, valueTextOut);
			}
		break;

		case eCTRL_CURSEL_AUTOMATION_MODE :
			if(tr)
			{
				//int mode = -1;
				//void* tmp = GetSetMediaTrackInfo(tr, "I_AUTOMODE", NULL);
				//if(tmp)
				//	mode = *(int*)(tmp);
				int mode = GetTrackAutomationMode(tr);
				if(valueTextOut)
					_device->getAutomationModeValueStr(mode, valueTextOut);
			}
		break;

		case eCTRL_SOLO :
		case eCTRL_CURSEL_SOLO :
			if(tr)
				value = boolStatus2Value(soloed, valueTextOut);
		break;

		case eCTRL_ARM :
		case eCTRL_CURSEL_ARM :
			if(tr)
				value = boolStatus2Value(recarmed, valueTextOut);
		break;

		case eCTRL_HZOOM_WHEEL :
		case eCTRL_VZOOM_WHEEL :
		case eCTRL_FFWRWD_WHEEL :
		case eCTRL_PREVNEXTMARKER_WHEEL :
		case eCTRL_VSCROLL_WHEEL :
		case eCTRL_HSCROLL_WHEEL :
		case eCTRL_ASSIGNABLEFNENC :
		case eCTRL_SCRUB_JOG_WHEEL :
		case eCTRL_TCPFXPARAM_ENC :
		case eCTRL_ITEM_PITCH_OCTAVE_WHEEL :
		case eCTRL_ITEM_PITCH_SEMITONE_WHEEL :
		case eCTRL_ITEM_PITCH_ONECENT_WHEEL :
		case eCTRL_TAKESELECT_WHEEL :
		case eCTRL_ITEMSELECT_WHEEL :
		case eCTRL_AUTOMATION_ACTIVE_FADER_WHEEL :
		case eCTRL_MOVE_CURSOR_GRID_WHEEL :
		case eCTRL_SPEED_DIAL :
			value = 0.5;
			if(valueTextOut)
				AutomapReaperClient::strcpyCrop(valueTextOut, "-/+");
		break;

		case eCTRL_HZOOM_POT :
		case eCTRL_VZOOM_POT :
		case eCTRL_VSCROLL_POT :
		case eCTRL_HSCROLL_POT :
		case eCTRL_SCRUB_JOG_POT :
			value = controller->getValue();
			if(valueTextOut)
				AutomapReaperClient::strcpyCrop(valueTextOut, "-/+");
		break;

		case eCTRL_TCPFXPARAM :
		{
			value = controller->getValue(controller->getSlot(idx));
			int midi = Val2Midi(value);
			if(valueTextOut)
				AutomapReaperClient::sprintfCrop(valueTextOut, "%03d", midi);
		}
		break;

		case eCTRL_TRACKSELECT_WHEEL :
			value = 0.5;
			if(tr && valueTextOut)
				AutomapReaperClient::getTrackNameCrop(tr, valueTextOut, AUTOMAP_NAME_LENGTH);
		break;

		case eCTRL_TRACKSELECT :
			value = (_parent->getSelTrack() == trackIndex);
			if(valueTextOut)
				_device->getTrackSelectValueStr(value, valueTextOut);
		break;

		case eCTRL_SUBMENUSELECT_WHEEL :
			if(AutomapRemoteSlDynamic* dev = dynamic_cast<AutomapRemoteSlDynamic*>(_device) )
			{
				value = 0.5;
				if(valueTextOut)
					dev->getSubmenuName(dev->getSubmenu(), valueTextOut);
			}
		break;

		case eCTRL_CURSEL_FXSELECT_WHEEL :
			value = 0.5;
			if( tr && valueTextOut )
			{
				char buffer[BUFFER_SIZE];
				bool res = TrackFX_GetFXName(tr, _parent->getSelFx(), buffer, AUTOMAP_NAME_LENGTH);
				if(res)
					AutomapReaperClient::strcpyCrop(valueTextOut, buffer);
				else
					AutomapReaperClient::strcpyCrop(valueTextOut, "<off>");
			}
		break;

		case eCTRL_CURSEL_FXPARAMSELECT_WHEEL :
		{
			value = 0.5;
			if( tr && valueTextOut )
			{
				char buffer[BUFFER_SIZE];
				bool res = TrackFX_GetParamName(tr, _parent->getSelFx(), _parent->getSelFxParam(), buffer, AUTOMAP_NAME_LENGTH);
				if(res)
					AutomapReaperClient::strcpyCrop(valueTextOut, buffer);
				else
					AutomapReaperClient::strcpyCrop(valueTextOut, "<off>");
			}
		}
		break;

		case eCTRL_CURSEL_TCPFXPARAMSELECT_WHEEL :
		{
			value = 0.5;
			if(valueTextOut)
				AutomapReaperClient::sprintfCrop(valueTextOut, "%02d", _parent->getSelTcpFxParam());
		}
		break;

		case eCTRL_CURSEL_TRACKSENDSELECT_WHEEL :
		{
			value = 0.5;
			if( tr && valueTextOut )
			{
				MediaTrack *dest = (MediaTrack*)GetSetTrackSendInfo(tr, 0, _parent->getSelTrackSend(), "P_DESTTRACK", NULL);
				if(dest)
					getTrackNameCrop(dest, valueTextOut, AUTOMAP_NAME_LENGTH);
				else
					AutomapReaperClient::strcpyCrop(valueTextOut, "<off>");
			}
		}
		break;

		case eCTRL_CURSEL_TRACKHARDSENDSELECT_WHEEL :
		{
			value = 0.5;
			if( tr && valueTextOut )
			{
				void* tmp = GetSetTrackSendInfo(tr, 1, _parent->getSelTrackHardSend(), "D_VOL", NULL);
				if(tmp)
					AutomapReaperClient::sprintfCrop(valueTextOut, "Out%d", _parent->getSelTrackHardSend()+1);
				else
					AutomapReaperClient::strcpyCrop(valueTextOut, "<off>");
			}
		}
		break;

		case eCTRL_FXPARAM :
			value = getFxParamValue(tr, controller->getType(), controller->getSlot(idx), valueTextOut);
		break;

		case eCTRL_TRACKSEND_VOL :
		case eCTRL_TRACKSEND_PAN :
		case eCTRL_TRACKSEND_MUTE :
		case eCTRL_TRACKSEND_MODE :
		case eCTRL_TRACKSEND_MUTE_POT :
		case eCTRL_TRACKSEND_MODE_POT :
		case eCTRL_TRACKHARDSEND_VOL :
		case eCTRL_TRACKHARDSEND_PAN :
		case eCTRL_TRACKHARDSEND_MUTE :
		case eCTRL_TRACKHARDSEND_MODE :
		case eCTRL_TRACKHARDSEND_MUTE_POT :
		case eCTRL_TRACKHARDSEND_MODE_POT :
			value = getTrackSendValue(tr, controller->getType(), controller->getSlot(idx), valueTextOut);
		break;

		case eCTRL_MCPMODE :
			value = 1.0*(_parent->getMcpMode());
			if(valueTextOut)
				_device->getMcpModeValueStr(value, valueTextOut);
		break;

		case eCTRL_CURSEL_FXBYPASS :
			if(tr)
				value = boolStatus2Value(fx, valueTextOut);
		break;

		case eCTRL_ASSIGNABLEFNPOT :
		case eCTRL_AUTOMATION_ACTIVE_FADER_POT :
			value = controller->getValue(controller->getSlot(idx));
			if(valueTextOut)
				AutomapReaperClient::sprintfCrop(valueTextOut, "%3d", Val2Midi(value));
		break;

		//case eCTRL_PLAY :
		//	value = 1.0*(playing && !pause);
		//	if(valueTextOut)
		//		_device->getButtonValueStr(value, valueTextOut);
		//break;
		//	
		//case eCTRL_RECORD :
		//	value = 1.0*(recording && !pause);
		//	if(valueTextOut)
		//		_device->getButtonValueStr(value, valueTextOut);
		//break;

		//case eCTRL_LOOP :
		//{
		//	int loop = GetSetRepeat(-1);
		//	value = 1.0*loop;
		//	if(valueTextOut)
		//		_device->getButtonValueStr(value, valueTextOut);
		//}
		//break;

		//case eCTRL_STOP :
		//	value = 1.0*(!playing && !pause && !recording);
		//	if(valueTextOut)
		//		_device->getButtonValueStr(value, valueTextOut);
		//break;

		case eMIDICTRL_VZOOM_WHEEL :
		case eMIDICTRL_HZOOM_WHEEL :
		case eMIDICTRL_VSCROLL_WHEEL :
		case eMIDICTRL_HSCROLL_WHEEL :
		case eMIDICTRL_VMOVE_WHEEL :
		case eMIDICTRL_HMOVE_WHEEL :
		case eMIDICTRL_VELOCITY_FINE_WHEEL :
		case eMIDICTRL_VELOCITY_COARSE_WHEEL :
		case eMIDICTRL_NOTE_LENGTH_WHEEL :
			value = 0.5;
			if(valueTextOut)
				AutomapReaperClient::strcpyCrop(valueTextOut, "-/+");
		break;

		case eMIDICTRL_GRID_WHEEL :
			value = 0.5;
			if(valueTextOut)
				_device->getGridValueStr((GridDivision)(_parent->getMidiGrid()), valueTextOut);
		break;

		case eMIDICTRL_TOGGLE_VIEWMODE :
			if(valueTextOut)
				_device->getMidiViewModeValueStr(_parent->getMidiViewMode(), valueTextOut);
		break;

		case eCTRL_ENVELOPE_DEFAULT_SHAPE :
			if(valueTextOut)
				_device->getEnvShapeValueStr(_parent->getDefaultEnvShape(), valueTextOut);
		break;

		case eCTRL_ENVELOPE_SELECTED_SHAPE :
			if(valueTextOut)
				_device->getEnvShapeValueStr(_parent->getSelectedEnvShape(), valueTextOut);
		break;

		case eMIDICTRL_CCLANE_POT :
		{
			value = controller->getValue();
			int lane = Val2Midi(value);
			if(valueTextOut)
				_device->getCcLaneValueStr(lane, valueTextOut);
		}
		break;

		case eCTRL_METRONOME_VOLUME		:
		{
			value = controller->getValue(controller->getSlot(idx));
			int midi = Val2Midi(value);
			if(valueTextOut)
				AutomapReaperClient::sprintfCrop(valueTextOut, "%03d", midi);
			double vol = value2Vol(value);
			vol2Value(vol, valueTextOut);
		}
		break;

		case eCTRL_SOLO_INFRONT_ADJUST_POT		:
		{
			value = controller->getValue(controller->getSlot(idx));
			double vol = -60.0*(1.0-value);
			if(valueTextOut)
				AutomapReaperClient::sprintfCrop(valueTextOut, "%+2.1lfdB", vol);
		}
		break;

		case eCTRL_GRID_WHEEL :
			value = 0.5;
			if(valueTextOut)
				_device->getGridValueStr((GridDivision)(_parent->getMainGrid()), valueTextOut);
		break;

		case eCTRL_FFW						:
		case eCTRL_RWD						:
		case eCTRL_PLAY						:
		case eCTRL_RECORD					:
		case eCTRL_STOP						:
		case eCTRL_LOOP						:
		case eCTRL_GOSTART					:
		case eCTRL_GOEND					:
		case eCTRL_ASSIGNABLEFN				:
		case eCTRL_MARKER_PREV				:
		case eCTRL_MARKER_NEXT				:
		case eCTRL_LOOP_SETSTART			:
		case eCTRL_LOOP_SETEND				:
		case eCTRL_INSERT_MARKERRGN			:
		case eCTRL_INSERT_MARKER			:
		case eCTRL_DELETE_MARKER			:
		case eCTRL_CLEARRECARM				:
		case eCTRL_CURSEL_NAME				:
		case eMIDICTRL_TOGGLE_SNAP			:
		case eMIDICTRL_TOGGLE_GRID			:
		case eMIDICTRL_MOVE_UP				:
		case eMIDICTRL_MOVE_DOWN			:
		case eMIDICTRL_MOVE_RIGHT			:
		case eMIDICTRL_MOVE_LEFT			:
		case eMIDICTRL_INSERTNOTE			:
		case eMIDICTRL_NOTEPROPERTIES		:
		case eMIDICTRL_NEXTSECONDARY		:
		case eMIDICTRL_VELOCITY_INC_FINE	:
		case eMIDICTRL_VELOCITY_INC_COARSE	:
		case eMIDICTRL_VELOCITY_DEC_FINE	:
		case eMIDICTRL_VELOCITY_DEC_COARSE	:
		case eMAINMIDICTRL_INSERTITEM		:
		case eMAINMIDICTRL_OPENITEMS		:
		case eMIDICTRL_TRANSPOSE_INC1		:
		case eMIDICTRL_TRANSPOSE_DEC1		:
		case eMIDICTRL_TRANSPOSE_INC12		:
		case eMIDICTRL_TRANSPOSE_DEC12		:
		case eMIDICTRL_HUMANIZE				:
		case eMIDICTRL_FIT_CONTENT			:
		case eMIDICTRL_FIT_LOOP				:
		case eMIDICTRL_DELETENOTE			:
		case eMIDICTRL_NOTE_MOVELEFT		:
		case eMIDICTRL_NOTE_MOVERIGHT		:
		case eMIDICTRL_NOTE_SELECTNEXT		:
		case eMIDICTRL_NOTE_SELECTPREV		:
		case eMIDICTRL_NOTE_VALUE_SELNEXT	:
		case eMIDICTRL_NOTE_VALUE_SELPREV	:
		case eCTRL_GROUPING_OVERRIDE		:
		case eCTRL_GRID_LINES				:
		case eCTRL_METRONOME				:
		case eCTRL_SNAP						:
		case eCTRL_MIXER_VISIBLE			:
		case eCTRL_TRACK_GROUPING_MATRIX	:
		case eCTRL_PERFORMANCE_METER		:
		case eCTRL_ROUTING_MATRIX			:
		case eCTRL_NAVIGATOR_WINDOW			:
		case eCTRL_SWS_TRACKLIST			:
		case eCTRL_TRACK_FREEPOS			:
		case eCTRL_ITEM_MOVE_DOWN			:
		case eCTRL_ITEM_MOVE_UP				:
		case eCTRL_ITEM_MOVE_RIGHT			:
		case eCTRL_ITEM_MOVE_LEFT			:
		case eCTRL_ITEM_GROUP					:
		case eCTRL_ITEM_SELECTALL_INGROUP		:
		case eCTRL_ITEM_REMOVE_FROMGROUP		:
		case eCTRL_ITEM_SELECTALL_INTIMESEL		:
		case eCTRL_ITEM_SELECTALL_INSELTRACK	:
		case eCTRL_ITEM_UNSELECTALL				:
		case eCTRL_ITEM_PITCH_ONECENT_DOWN		:
		case eCTRL_ITEM_PITCH_OCTAVE_DOWN		:
		case eCTRL_ITEM_PITCH_SEMITONE_DOWN		:
		case eCTRL_ITEM_PITCH_ONECENT_UP		:
		case eCTRL_ITEM_PITCH_OCTAVE_UP			:
		case eCTRL_ITEM_PITCH_SEMITONE_UP		:
		case eCTRL_ITEM_SELECT_NEXTRACK			:
		case eCTRL_ITEM_SELECT_PREVTRACK		:
		case eCTRL_ITEM_SELECT_NEXTITEM			:
		case eCTRL_ITEM_SELECT_PREVITEM			:
		case eCTRL_TAKE_ACTIVE_CROP				:
		case eCTRL_TAKE_ACTIVE_DELETE			:
		case eCTRL_TAKE_ACTIVE_NEXT				:
		case eCTRL_TAKE_ACTIVE_PREVIOUS			:
		//! \todo Use item "B_MUTE" to get state
		case eCTRL_ITEM_MUTE					:
		//! \todo Use item "B_LOOPSRC" to get state
		case eCTRL_ITEM_LOOP_SOURCE				:
		case eCTRL_ITEM_LOOP_SECTION			:
		case eCTRL_ITEM_PROPERTIES				:
		case eCTRL_CURSEL_VIEW_ENVELOPES		:
		case eCTRL_CURSEL_VIEW_FX	:
		case eCTRL_CURSEL_VIEW_IO	:
		case eCTRL_CURSEL_VIEW_GROUPING	:
		case eCTRL_TAKE_REVERSE	:
		case eCTRL_SHOW_ALL_ACTIVE_ENVELOPES	:
		case eCTRL_ENVELOPE_BYPASS	:
		case eCTRL_ENVELOPE_RECARM	:
		case eCTRL_ENVELOPE_INSERT_POINT		:
		case eCTRL_ENVELOPE_VISIBLE_VOLUME		:
		case eCTRL_ENVELOPE_VISIBLE_PAN			:
		case eCTRL_ENVELOPE_VISIBLE_PREFXVOL	:
		case eCTRL_ENVELOPE_VISIBLE_PREFXPAN	:
		case eCTRL_ENVELOPE_CLEAR	:
		case eCTRL_ENVELOPE_DELETE_POINTS	:
		case eCTRL_ENVELOPE_TOGGLE_DISPLAY	:
		case eCTRL_ENVELOPE_ALLVISIBLE_LANE	:
		case eCTRL_UNDO_HISTORY	:
		case eCTRL_SWS_MARKERLIST	:
		case eCTRL_SWS_SNAPSHOTLIST	:
		case eCTRL_SWS_ZOOM_TO_TRACKTIMESELECT	:
		case eCTRL_SWS_NEW_SNAPSHOT	:
		case eCTRL_SWS_SAVE_SNAPSHOT	:
		case eCTRL_ITEM_GLUED_SELECTED :
		case eCTRL_SHOW_ACTIONLIST :
		default :
			value = 0.0;
			if(valueTextOut)
				AutomapReaperClient::strcpyCrop(valueTextOut, " ");
		break;
	}

	return (float)(value);
}

void AutomapReaperClient::SetParamValue(int idx, float value)
{
	AutomapController* controller = _device->getController(idx);
	if(controller == NULL)
		return;

	int trackIndex = controller->getTrackIndex();
	if(trackIndex<0)
		trackIndex = _parent->getSelTrack();
	MediaTrack* tr = CSurf_TrackFromID(trackIndex, _parent->getMcpMode());

	int flags = GetPlayState();
	bool play = ((flags&1) != 0);
	bool pause = ((flags&2) != 0);
	bool rec = ((flags&4) != 0);

	flags = 0;
	if(tr)
		GetTrackInfo((INT_PTR)(tr), &flags);
	bool soloed = ((flags&16) != 0);
	bool muted = ((flags&8) != 0);
	bool recarmed = ((flags&64) != 0);
	bool fx = ((flags&4) != 0);

	double vald = value;

	switch(controller->getType())
	{
		case eCTRL_VOL :
		case eCTRL_CURSEL_VOL :
			if(tr)
				CSurf_SetSurfaceVolume(tr,CSurf_OnVolumeChangeEx(tr,value2Vol(value),false,false),NULL);
		break;

		case eCTRL_PAN :
		case eCTRL_CURSEL_PAN :
			if(tr)
				CSurf_SetSurfacePan(tr,CSurf_OnPanChangeEx(tr,value2Pan(value),false,false),NULL);
		break;

		case eCTRL_MUTE :
		case eCTRL_CURSEL_MUTE :
			if(tr)
				CSurf_SetSurfaceMute(tr,CSurf_OnMuteChange(tr,1*(!muted)),NULL);
		break;

		case eCTRL_PHASE :
		case eCTRL_CURSEL_PHASE :
			if(tr)
			{
				void* tmp = GetSetMediaTrackInfo(tr, "B_PHASE", NULL);
				if(tmp)
				{
					bool phased = ! *(bool*)(tmp);
					GetSetMediaTrackInfo(tr, "B_PHASE", &phased);
				}
			}
		break;

		case eCTRL_CURSEL_AUTOMATION_MODE :
			if(tr)
			{
				int mode = GetTrackAutomationMode(tr);
				mode++;
				mode = mode%5;
				SetTrackAutomationMode(tr, mode);
				CSurf_SetAutoMode(mode, this->_parent);
				//void* tmp = GetSetMediaTrackInfo(tr, "I_AUTOMODE", NULL);
				//if(tmp)
				//{
				//	int mode = *(int*)(tmp);
				//	mode++;
				//	mode = mode%5;
				//	GetSetMediaTrackInfo(tr, "I_AUTOMODE", &mode);
				//}
			}
		break;


		case eCTRL_SOLO :
		case eCTRL_CURSEL_SOLO :
			if(tr)
				CSurf_SetSurfaceSolo(tr,CSurf_OnSoloChange(tr,1*(!soloed)),NULL);
		break;

		case eCTRL_ARM :
		case eCTRL_CURSEL_ARM :
			if(tr)
				CSurf_SetSurfaceRecArm(tr,CSurf_OnRecArmChange(tr,1*(!recarmed)),NULL);
//! \bug BR: Cockos: Does track RecArm call TrackList_UpdateAllExternalSurfaces()? Why?
//return;		// Controller will be erased by map refresh: you want to skip controller->isEndless()
		break;

		case eCTRL_CURSEL_FXBYPASS :
		if(tr)
			CSurf_OnFXChange(tr, 1*(!fx));
		break;

		case eCTRL_HZOOM_WHEEL :
		{
			if(value<0.5)
				adjustZoom(-_parent->getZoomFactor(), 0, true, -1);
			else
				adjustZoom(_parent->getZoomFactor(), 0, true, -1);
		}
		break;

		case eCTRL_VZOOM_WHEEL :
		{
			if(value<0.5)
				Main_OnCommand(ID_MAIN_VZOOM_OUT, 0);
			else
				Main_OnCommand(ID_MAIN_VZOOM_IN, 0);
		}
		break;

		case eCTRL_TRACKSELECT_WHEEL :
		{
			int trackIndex = _parent->getSelTrack();
			trackIndex += 2*(value>=0.5) - 1;
			_parent->selectTrack(trackIndex);
		}
		break;

		case eCTRL_TRACKSELECT :
			_parent->selectTrack(trackIndex);
		break;

		case eCTRL_CURSEL_NAME :
		if(_parent->_automapRawMidiClient)
		{
			_parent->_automapRawMidiClient->clearPanel(AutomapSLRawMidiClient::ePANEL_LEFT);
			_parent->_automapRawMidiClient->sendText(AutomapSLRawMidiClient::ePANEL_TOPLEFT, "t:Name:", 1);
			if(tr)
			{
				const char* fulltitle = GetTrackInfo((INT_PTR)(tr), NULL);
				_parent->_automapRawMidiClient->sendText(AutomapSLRawMidiClient::ePANEL_TOPLEFT, fulltitle, 9);
			}
			else
				_parent->_automapRawMidiClient->sendText(AutomapSLRawMidiClient::ePANEL_TOPLEFT, "???", 9);
		}
		break;

		case eCTRL_SUBMENUSELECT_WHEEL :
			if(AutomapRemoteSlDynamic* dev = dynamic_cast<AutomapRemoteSlDynamic*>(_device) )
			{
				int submenu = dev->getSubmenu();
				submenu += 2*(value>=0.5) - 1;
				if(submenu<0)
					submenu = 0;
				if(submenu>(dev->getNbSubmenus()-1))
					submenu = dev->getNbSubmenus()-1;
				dev->setSubmenu(submenu);

				//int submenu = (int)(dev->getSubmenu());
				//submenu += 2*(value>=0.5) - 1;
				//if(submenu<0)
				//	submenu = 0;
				//if(submenu>(int)(AutomapRemoteSlDynamic::eNBSUBMENUS)-1)
				//	submenu = (int)(AutomapRemoteSlDynamic::eNBSUBMENUS)-1;
				//dev->setSubmenu((AutomapRemoteSlDynamic::Submenu)(submenu));
				_parent->resetMapping(-1);
				return;		// Controller will be erased: you want to skip controller->isEndless()
			}
		break;

		case eCTRL_PREVNEXTMARKER_WHEEL :
			if(value<0.5)
				Main_OnCommand(ID_MARKER_PREV, 0);
			else
				Main_OnCommand(ID_MARKER_NEXT, 0);
		break;

		case eCTRL_CURSEL_FXSELECT_WHEEL :
		{
			int selFx = _parent->getSelFx();
			selFx += 2*(value>=0.5) - 1;
			_parent->selectFx(tr, selFx);
		}
		break;

		case eCTRL_CURSEL_FXPARAMSELECT_WHEEL :
		{
			int selFxParam = _parent->getSelFxParam();
			selFxParam += 2*(value>=0.5) - 1;
			_parent->selectFxParam(tr, selFxParam);
		}
		break;

		case eCTRL_CURSEL_TCPFXPARAMSELECT_WHEEL :
		{
			int selTcpFxParam = _parent->getSelTcpFxParam();
			selTcpFxParam += 2*(value>=0.5) - 1;
			_parent->selectTcpFxParam(selTcpFxParam);
		}
		break;

		case eCTRL_CURSEL_TRACKSENDSELECT_WHEEL :
		{
			int selTrackSend = _parent->getSelTrackSend();
			selTrackSend += 2*(value>=0.5) - 1;
			_parent->selectTrackSend(tr, selTrackSend);
		}
		break;

		case eCTRL_CURSEL_TRACKHARDSENDSELECT_WHEEL :
		{
			int selTrackHardSend = _parent->getSelTrackHardSend();
			selTrackHardSend += 2*(value>=0.5) - 1;
			_parent->selectHardTrackSend(tr, selTrackHardSend);
		}
		break;

		case eCTRL_FXPARAM :
			setFxParamValue(tr, controller->getType(), controller->getSlot(idx), (double)(value));
		break;

		case eCTRL_TRACKSEND_VOL :
		case eCTRL_TRACKSEND_PAN :
		case eCTRL_TRACKSEND_MUTE :
		case eCTRL_TRACKSEND_MODE :
		case eCTRL_TRACKSEND_MUTE_POT :
		case eCTRL_TRACKSEND_MODE_POT :
		case eCTRL_TRACKHARDSEND_VOL :
		case eCTRL_TRACKHARDSEND_PAN :
		case eCTRL_TRACKHARDSEND_MUTE :
		case eCTRL_TRACKHARDSEND_MODE :
		case eCTRL_TRACKHARDSEND_MUTE_POT :
		case eCTRL_TRACKHARDSEND_MODE_POT :
			setTrackSendValue(tr, controller->getType(), controller->getSlot(idx), (double)(value));
		break;

		case eCTRL_PLAY :
			CSurf_OnPlay();
			//if(play && !rec)
			//	CSurf_SetPlayState(false, true, false, NULL);
			//else if(play && rec)
			//	CSurf_SetPlayState(true, false, false, NULL);
			//else if(!play)
			//	CSurf_SetPlayState(true, false, false, NULL);
		break;

		case eCTRL_STOP:
			CSurf_OnStop();
			//CSurf_SetPlayState(false, false, false, NULL);
		break;

		case eCTRL_RECORD :
			CSurf_OnRecord();
			//if(!rec)
			//	CSurf_SetPlayState(true, false, true, NULL);
			//else
			//	CSurf_SetPlayState(true, false, false, NULL);
		break;

		case eCTRL_FFW :
			setFfwRew(-1, AutomapCSurf::Parameters::eFFWRWD_MOVECURSOR, true);
		break;

		case eCTRL_RWD :
			setFfwRew(1, AutomapCSurf::Parameters::eFFWRWD_MOVECURSOR, true);
		break;

		case eCTRL_FFWRWD_WHEEL :
			if(value<0.5)
				MoveEditCursor(-_parent->getFfwRwdRange(), false);
				//CSurf_ScrubAmt(-100.0*_parent->getFfwRwdRange());
			else
				MoveEditCursor(_parent->getFfwRwdRange(), false);
				//CSurf_ScrubAmt(100.0*_parent->getFfwRwdRange());
		break;

		case eCTRL_VSCROLL_WHEEL :
			CSurf_OnArrow(1*(value>=0.5), false);
		break;

		case eCTRL_HSCROLL_WHEEL :
			if(value<0.5)
				Main_OnCommand(ID_MAIN_SCROLL_LEFT, 0);
			else
				Main_OnCommand(ID_MAIN_SCROLL_RIGHT, 0);
		break;

		case eCTRL_GOSTART :
			CSurf_GoStart();
		break;

		case eCTRL_GOEND :
			CSurf_GoEnd();
		break;

		case eCTRL_LOOP :
			GetSetRepeat(2);	// Toggle
		break;

		case eMAINMIDICTRL_INSERTITEM :
			Main_OnCommand(ID_MAIN_MIDI_INSERT_ITEM, 0);
		break;

		case eMAINMIDICTRL_OPENITEMS :
			Main_OnCommand(ID_MAIN_MIDI_OPEN_ITEMS, 0);
		break;

		case eMIDICTRL_VZOOM_WHEEL :
			if(value<0.5)
				MidiEdit_OnCommand(ID_MIDI_VZOOM_OUT, 0);
			else
				MidiEdit_OnCommand(ID_MIDI_VZOOM_IN, 0);
		break;
		case eMIDICTRL_HZOOM_WHEEL :
			if(value<0.5)
				MidiEdit_OnCommand(ID_MIDI_HZOOM_OUT, 0);
			else
				MidiEdit_OnCommand(ID_MIDI_HZOOM_IN, 0);
		break;
		case eMIDICTRL_VSCROLL_WHEEL :
			if(value<0.5)
				MidiEdit_OnCommand(ID_MIDI_SCROLLVIEW_UP, 0);
			else
				MidiEdit_OnCommand(ID_MIDI_SCROLLVIEW_DOWN, 0);
		break;

		case eMIDICTRL_HSCROLL_WHEEL :
			//! \bug Cockos: can't make KBD_OnActionEx() work with MIDI Editor window
			//if(value<0.5)
			//	KBD_OnMidiEditActionEx(ID_MIDI_CCONLY_HSCROLL, 0x3F, -1, 2, NULL);
			//else
			//	KBD_OnMidiEditActionEx(ID_MIDI_CCONLY_HSCROLL, 0x41, -1, 2, NULL);
			if(value<0.5)
				MidiEdit_OnCommand(ID_MIDI_SCROLLVIEW_LEFT, 0);
			else
				MidiEdit_OnCommand(ID_MIDI_SCROLLVIEW_RIGHT, 0);
		break;
		case eMIDICTRL_HMOVE_WHEEL :
			if(value<0.5)
				MidiEdit_OnCommand(ID_MIDI_MOVE_LEFT, 0);
			else
				MidiEdit_OnCommand(ID_MIDI_MOVE_RIGHT, 0);
		break;
		case eMIDICTRL_VMOVE_WHEEL :
			if(value<0.5)
				MidiEdit_OnCommand(ID_MIDI_MOVE_DOWN, 0);
			else
				MidiEdit_OnCommand(ID_MIDI_MOVE_UP, 0);
		break;

		case eMIDICTRL_VELOCITY_COARSE_WHEEL :
			if(value<0.5)
				MidiEdit_OnCommand(ID_MIDI_VELOCITY_DEC10, 0);
			else
				MidiEdit_OnCommand(ID_MIDI_VELOCITY_INC10, 0);
		break;

		case eMIDICTRL_VELOCITY_FINE_WHEEL :
			if(value<0.5)
				MidiEdit_OnCommand(ID_MIDI_VELOCITY_DEC1, 0);
			else
				MidiEdit_OnCommand(ID_MIDI_VELOCITY_INC1, 0);
		break;

		case eMIDICTRL_GRID_WHEEL :
		{
			int grid = _parent->getMidiGrid();
			grid += 2*(value>=0.5) - 1;
			_parent->selectMidiGrid(grid);
		}
		break;

		case eMIDICTRL_NOTE_LENGTH_WHEEL :
		{
			if(value<0.5)
				MidiEdit_OnCommand(ID_MIDI_NOTE_SHORTEN, 0);
			else
				MidiEdit_OnCommand(ID_MIDI_NOTE_LENGTHEN, 0);
		}
		break;

		case eMIDICTRL_TOGGLE_VIEWMODE :
		{
			int view = _parent->getMidiViewMode();
			view++;
			view = view%3;
			_parent->setMidiViewMode(view);
			MidiEdit_OnCommand(_device->getMidiViewModeCommand(view), 0);
		}
		break;

		case eCTRL_ENVELOPE_DEFAULT_SHAPE :
		{
			int shape = _parent->getDefaultEnvShape();
			shape++;
			shape = shape%6;
			_parent->setDefaultEnvShape(shape);
			Main_OnCommand(_device->getDefaultEnvShapeCommand(shape), 0);
		}
		break;

		case eCTRL_ENVELOPE_SELECTED_SHAPE :
		{
			int shape = _parent->getSelectedEnvShape();
			shape++;
			shape = shape%6;
			_parent->setSelectedEnvShape(shape);
			Main_OnCommand(_device->getSelectedEnvShapeCommand(shape), 0);
		}
		break;

		case eMIDICTRL_TOGGLE_SNAP :
			MidiEdit_OnCommand(ID_MIDI_TOGGLESNAP, 0);
		break;
		case eMIDICTRL_TOGGLE_GRID :
			MidiEdit_OnCommand(ID_MIDI_TOGGLEGRID, 0);
		break;
		case eMIDICTRL_MOVE_UP :
			MidiEdit_OnCommand(ID_MIDI_MOVE_UP, 0);
		break;
		case eMIDICTRL_MOVE_DOWN :
			MidiEdit_OnCommand(ID_MIDI_MOVE_DOWN, 0);
		break;
		case eMIDICTRL_MOVE_RIGHT :
			MidiEdit_OnCommand(ID_MIDI_MOVE_RIGHT, 0);
		break;
		case eMIDICTRL_MOVE_LEFT :
			MidiEdit_OnCommand(ID_MIDI_MOVE_LEFT, 0);
		break;
		case eMIDICTRL_INSERTNOTE :
			MidiEdit_OnCommand(ID_MIDI_UNSELECT_ALL, 0);
			MidiEdit_OnCommand(ID_MIDI_INSERTNOTE_EDITCURSOR, 0);
		break;
		case eMIDICTRL_NOTEPROPERTIES :
			MidiEdit_OnCommand(ID_MIDI_NOTEPROPERTIES, 0);
		break;
		case eMIDICTRL_NEXTSECONDARY :
			MidiEdit_OnCommand(ID_MIDI_NEXT_SECONDARY, 0);
		break;
		case eMIDICTRL_VELOCITY_INC_FINE :
			MidiEdit_OnCommand(ID_MIDI_VELOCITY_INC1, 0);
		break;
		case eMIDICTRL_VELOCITY_INC_COARSE :
			MidiEdit_OnCommand(ID_MIDI_VELOCITY_INC10, 0);
		break;
		case eMIDICTRL_VELOCITY_DEC_FINE :
			MidiEdit_OnCommand(ID_MIDI_VELOCITY_DEC1, 0);
		break;
		case eMIDICTRL_VELOCITY_DEC_COARSE :
			MidiEdit_OnCommand(ID_MIDI_VELOCITY_DEC10, 0);
		break;
		case eMIDICTRL_TRANSPOSE_INC1	:
			MidiEdit_OnCommand(ID_MIDI_TRANSPOSE_INC1, 0);
		break;
		case eMIDICTRL_TRANSPOSE_DEC1	:
			MidiEdit_OnCommand(ID_MIDI_TRANSPOSE_DEC1, 0);
		break;
		case eMIDICTRL_TRANSPOSE_INC12	:
			MidiEdit_OnCommand(ID_MIDI_TRANSPOSE_INC12, 0);
		break;
		case eMIDICTRL_TRANSPOSE_DEC12	:
			MidiEdit_OnCommand(ID_MIDI_TRANSPOSE_DEC12, 0);
		break;
		case eMIDICTRL_HUMANIZE			:
			MidiEdit_OnCommand(ID_MIDI_HUMANIZE, 0);
		break;
		case eMIDICTRL_FIT_CONTENT		:
			MidiEdit_OnCommand(ID_MIDI_FIT_CONTENT, 0);
		break;
		case eMIDICTRL_FIT_LOOP			:
			MidiEdit_OnCommand(ID_MIDI_FIT_LOOP, 0);
		break;
		case eMIDICTRL_DELETENOTE			:
			MidiEdit_OnCommand(ID_MIDI_DELETENOTE, 0);
		break;
		case eMIDICTRL_NOTE_MOVELEFT			:
			MidiEdit_OnCommand(ID_MAIN_MIDI_NOTE_MOVELEFT, 0);
		break;
		case eMIDICTRL_NOTE_MOVERIGHT			:
			MidiEdit_OnCommand(ID_MAIN_MIDI_NOTE_MOVERIGHT, 0);
		break;
		case eMIDICTRL_NOTE_SELECTNEXT			:
			MidiEdit_OnCommand(ID_MAIN_MIDI_NOTE_SELECTNEXT, 0);
		break;
		case eMIDICTRL_NOTE_SELECTPREV			:
			MidiEdit_OnCommand(ID_MAIN_MIDI_NOTE_SELECTPREV, 0);
		break;
		case eMIDICTRL_NOTE_VALUE_SELNEXT			:
			MidiEdit_OnCommand(ID_MAIN_MIDI_NOTE_VALUE_SELNEXT, 0);
		break;
		case eMIDICTRL_NOTE_VALUE_SELPREV			:
			MidiEdit_OnCommand(ID_MAIN_MIDI_NOTE_VALUE_SELPREV, 0);
		break;

		case eMIDICTRL_CCLANE_POT :
		{
			controller->setValue(value);
			int lane = Val2Midi(value);
			_parent->selectMidiCcLane(lane);
		}
		break;

		case eCTRL_HZOOM_POT :
		{
			controller->setValue(value);
			int midi = Val2Midi(value);
			KBD_OnMainActionEx(ID_MAIN_CCONLY_ADJUST_HORIZONTAL_ZOOM, midi, -1, 0, NULL, NULL);
		}
		break;

		case eCTRL_VZOOM_POT :
		{
			controller->setValue(value);
			int midi = Val2Midi(value);
			KBD_OnMainActionEx(ID_MAIN_CCONLY_ADJUST_VERTICAL_ZOOM, midi, -1, 0, NULL, NULL);
		}
		break;

		case eCTRL_HSCROLL_POT :
		{
			controller->setValue(value);
			int midi = Val2Midi(value);
			KBD_OnMainActionEx(ID_MAIN_CCONLY_ADJUST_HORIZONTAL_SCROLL, midi, -1, 0, NULL, NULL);
		}
		break;

		case eCTRL_VSCROLL_POT :
		{
			controller->setValue(value);
			int midi = Val2Midi(value);
			KBD_OnMainActionEx(ID_MAIN_CCONLY_ADJUST_VERTICAL_SCROLL, midi, -1, 0, NULL, NULL);
		}
		break;

		case eCTRL_SCRUB_JOG_POT :
		{
			controller->setValue(value);
			int midi = Val2Midi(value);
			KBD_OnMainActionEx(ID_MAIN_CCONLY_SCRUB_JOG, midi, -1, 0, NULL, NULL);
		}
		break;

		case eCTRL_SCRUB_JOG_WHEEL :
		{
			if(value<0.5)
				KBD_OnMainActionEx(ID_MAIN_CCONLY_SCRUB_JOG, 0x3F, -1, 2, NULL, NULL);
			else
				KBD_OnMainActionEx(ID_MAIN_CCONLY_SCRUB_JOG, 0x41, -1, 2, NULL, NULL);
		}
		break;

		case eCTRL_TCPFXPARAM :
		{
			controller->setValue(value, controller->getSlot(idx));
			int midi = Val2Midi(value);
			int cmd = ID_MAIN_CCONLY_SELTRACK_FXPARAM01+_parent->getSelTcpFxParam()+controller->getSlot(idx);
			KBD_OnMainActionEx(cmd, midi, -1, 0, NULL, NULL);
		}
		break;

		case eCTRL_TCPFXPARAM_ENC :
		{
			int cmd = ID_MAIN_CCONLY_SELTRACK_FXPARAM01+_parent->getSelTcpFxParam()+controller->getSlot(idx);
			if(value<0.5)
				KBD_OnMainActionEx(cmd, 0x3F, -1, 2, NULL, NULL);
			else
				KBD_OnMainActionEx(cmd, 0x41, -1, 2, NULL, NULL);
		}
		break;

		case eCTRL_ASSIGNABLEFN :
		{
			int cc = controller->getSlot(idx);
			int chan = 0xb0 | _parent->getFnMidiChannel();
			MIDI_event_t evt={0,3,{chan,cc,0}};
			kbd_OnMidiEvent(&evt,-1);
		}
		break;

		case eCTRL_ASSIGNABLEFNPOT :
		{
			int cc = controller->getSlot(idx) + 64;
			controller->setValue(value, controller->getSlot(idx));
			int valint = Val2Midi(value);
			int chan = 0xb0 | _parent->getFnMidiChannel();
			MIDI_event_t evt={0,3,{chan,cc,valint}};
			kbd_OnMidiEvent(&evt,-1);
		}
		break;

		case eCTRL_ASSIGNABLEFNENC :
		{
			int cc = controller->getSlot(idx) + 96;
			int valint = 0x40;
			if(value<0.5)
				valint = 0x3F;
			else
				valint = 0x41;
			int chan = 0xb0 | _parent->getFnMidiChannel();
			MIDI_event_t evt={0,3,{chan,cc,valint}};
			kbd_OnMidiEvent(&evt,-1);
		}
		break;

		case eCTRL_MARKER_PREV :
			//SendMessage(g_hwnd, WM_COMMAND, ID_MARKER_PREV, 0);
			Main_OnCommand(ID_MARKER_PREV, 0);
		break;
		case eCTRL_MARKER_NEXT :
			Main_OnCommand(ID_MARKER_NEXT, 0);
		break;
		case eCTRL_LOOP_SETSTART :
			Main_OnCommand(ID_LOOP_SETSTART, 0);
		break;
		case eCTRL_LOOP_SETEND :
			Main_OnCommand(ID_LOOP_SETEND, 0);
		break;
		case eCTRL_INSERT_MARKERRGN :
			Main_OnCommand(ID_INSERT_MARKERRGN, 0);
		break;
		case eCTRL_INSERT_MARKER :
			Main_OnCommand(ID_MAIN_INSERT_EDIT_MARKER, 0);
		break;
		case eCTRL_DELETE_MARKER :
			Main_OnCommand(ID_MAIN_MARKER_DELETE, 0);
		break;

		case eCTRL_SHOW_ACTIONLIST :
			Main_OnCommand(ID_MAIN_SHOW_ACTIONLIST, 0);
		break;

		case eCTRL_ITEM_GLUED_SELECTED :
			Main_OnCommand(ID_MAIN_ITEM_GLUED_SELECTED, 0);
		break;

		case eCTRL_MCPMODE :
			_parent->toggleMcpMode();
			return;		// Controller will be erased: you want to skip controller->isEndless()
		break;

		case eCTRL_CLEARRECARM :
			ClearAllRecArmed();
		break;

		case eCTRL_GRID_WHEEL :
		{
			int grid = _parent->getMainGrid();
			grid += 2*(value>=0.5) - 1;
			_parent->selectMainGrid(grid);
		}
		break;

		case eCTRL_SPEED_DIAL :
		{
			if(value<0.5)
				mouse_event(MOUSEEVENTF_WHEEL, 0, 0, -_parent->getSpeedDialFactor(), 0);
			else
				mouse_event(MOUSEEVENTF_WHEEL, 0, 0, +_parent->getSpeedDialFactor(), 0);
		}
		break;

		case eCTRL_METRONOME_VOLUME :
		{
			controller->setValue(value);
			int midi = Val2Midi(value);
			KBD_OnMainActionEx(ID_MAIN_CCONLY_METRONOME_VOLUME, midi, -1, 0, NULL, NULL);
		}
		break;

		case eCTRL_SOLO_INFRONT_ADJUST_POT :
		{
			controller->setValue(value);
			int midi = Val2Midi(value);
			KBD_OnMainActionEx(ID_MAIN_SOLO_INFRONT_ADJUST_CCONLY, midi, -1, 0, NULL, NULL);
		}
		break;

		case eCTRL_GROUPING_OVERRIDE :
			Main_OnCommand(ID_MAIN_OPTION_TOGGLE_GROUPING_OVERRIDE, 0);
		break;
		case eCTRL_GRID_LINES :
			Main_OnCommand(ID_MAIN_OPTION_TOGGLE_GRID_LINES, 0);
		break;
		case eCTRL_METRONOME :
			Main_OnCommand(ID_MAIN_OPTION_TOGGLE_METRONOME, 0);
		break;
		case eCTRL_SNAP :
			Main_OnCommand(ID_MAIN_OPTION_TOGGLE_SNAP, 0);
		break;
		case eCTRL_MIXER_VISIBLE :
			Main_OnCommand(ID_MAIN_VIEW_TOGGLE_MIXER_VISIBLE, 0);
		break;
		case eCTRL_TRACK_GROUPING_MATRIX :
			Main_OnCommand(ID_MAIN_VIEW_SHOW_TRACK_GROUPING_MATRIX, 0);
		break;
		case eCTRL_PERFORMANCE_METER :
			Main_OnCommand(ID_MAIN_VIEW_SHOW_PERFORMANCE_METER, 0);
		break;
		case eCTRL_ROUTING_MATRIX :
			Main_OnCommand(ID_MAIN_VIEW_SHOW_ROUTING_MATRIX, 0);
		break;
		case eCTRL_NAVIGATOR_WINDOW :
			Main_OnCommand(ID_MAIN_VIEW_SHOW_NAVIGATOR_WINDOW, 0);
		break;

		case eCTRL_SWS_TRACKLIST :
			Main_OnExtendedCommand(ID_SWS_MAIN_SHOW_TRACKLIST, 0);
		break;
		case eCTRL_SWS_MARKERLIST :
			Main_OnExtendedCommand(ID_SWS_MAIN_SHOW_MARKERLIST, 0);
		break;
		case eCTRL_SWS_SNAPSHOTLIST :
			Main_OnExtendedCommand(ID_SWS_MAIN_SHOW_SNAPSHOTLIST, 0);
		break;
		case eCTRL_SWS_ZOOM_TO_TRACKTIMESELECT :
			Main_OnExtendedCommand(ID_SWS_MAIN_TOGGLE_ZOOM_SELTRACK_TIME, 0);
			return;
			//! \bug BR: SWS: Does action 'SWS: toggle zoom to select' call TrackList_UpdateAllExternalSurfaces()? Why?
		break;

		case eCTRL_SWS_NEW_SNAPSHOT :
			Main_OnExtendedCommand(ID_SWS_MAIN_NEW_SNAPSHOT_NAMED, 0);
		break;
		case eCTRL_SWS_SAVE_SNAPSHOT :
			Main_OnExtendedCommand(ID_SWS_MAIN_SAVE_SNAPSHOT, 0);
		break;

		case eCTRL_TRACK_FREEPOS :
			Main_OnCommand(ID_MAIN_TRACK_TOGGLE_FREEPOS, 0);
		break;
		case eCTRL_ITEM_MOVE_DOWN :
			Main_OnCommand(ID_MAIN_ITEM_MOVE_ONETRACK_DOWN, 0);
		break;
		case eCTRL_ITEM_MOVE_UP :
			Main_OnCommand(ID_MAIN_ITEM_MOVE_ONETRACK_UP, 0);
		break;
		case eCTRL_ITEM_MOVE_RIGHT :
			Main_OnCommand(ID_MAIN_ITEM_MOVE_GRIDSIZE_RIGHT, 0);
		break;
		case eCTRL_ITEM_MOVE_LEFT :
			Main_OnCommand(ID_MAIN_ITEM_MOVE_GRIDSIZE_LEFT, 0);
		break;
		case eCTRL_ITEM_GROUP :
			Main_OnCommand(ID_MAIN_ITEM_GROUP, 0);
		break;
		case eCTRL_ITEM_SELECTALL_INGROUP :
			Main_OnCommand(ID_MAIN_ITEM_SELECTALL_INGROUP, 0);
		break;
		case eCTRL_ITEM_REMOVE_FROMGROUP :
			Main_OnCommand(ID_MAIN_ITEM_REMOVE_FROMGROUP, 0);
		break;
		case eCTRL_ITEM_SELECTALL_INTIMESEL :
			Main_OnCommand(ID_MAIN_ITEM_SELECTALL_INTIMESEL, 0);
		break;
		case eCTRL_ITEM_SELECTALL_INSELTRACK :
			Main_OnCommand(ID_MAIN_ITEM_SELECTALL_INSELTRACK, 0);
		break;
		case eCTRL_ITEM_UNSELECTALL :
			Main_OnCommand(ID_MAIN_ITEM_UNSELECTALL, 0);
		break;
		case eCTRL_ITEM_PITCH_ONECENT_DOWN :
			Main_OnCommand(ID_MAIN_ITEM_PITCH_ONECENT_DOWN, 0);
		break;
		case eCTRL_ITEM_PITCH_OCTAVE_DOWN :
			Main_OnCommand(ID_MAIN_ITEM_PITCH_OCTAVE_DOWN, 0);
		break;
		case eCTRL_ITEM_PITCH_SEMITONE_DOWN :
			Main_OnCommand(ID_MAIN_ITEM_PITCH_SEMITONE_DOWN, 0);
		break;
		case eCTRL_ITEM_PITCH_ONECENT_UP :
			Main_OnCommand(ID_MAIN_ITEM_PITCH_ONECENT_UP, 0);
		break;
		case eCTRL_ITEM_PITCH_OCTAVE_UP :
			Main_OnCommand(ID_MAIN_ITEM_PITCH_OCTAVE_UP, 0);
		break;
		case eCTRL_ITEM_PITCH_SEMITONE_UP :
			Main_OnCommand(ID_MAIN_ITEM_PITCH_SEMITONE_UP, 0);
		break;
		case eCTRL_ITEM_SELECT_NEXTRACK :
			Main_OnCommand(ID_MAIN_ITEM_SELECT_NEXTRACK, 0);
		break;
		case eCTRL_ITEM_SELECT_PREVTRACK :
			Main_OnCommand(ID_MAIN_ITEM_SELECT_PREVTRACK, 0);
		break;
		case eCTRL_ITEM_SELECT_NEXTITEM :
			Main_OnCommand(ID_MAIN_ITEM_SELECT_NEXTITEM, 0);
		break;
		case eCTRL_ITEM_SELECT_PREVITEM :
			Main_OnCommand(ID_MAIN_ITEM_SELECT_PREVITEM, 0);
		break;
		case eCTRL_TAKE_ACTIVE_CROP :
			Main_OnCommand(ID_MAIN_TAKE_ACTIVE_CROP, 0);
		break;
		case eCTRL_TAKE_ACTIVE_DELETE :
			Main_OnCommand(ID_MAIN_TAKE_ACTIVE_DELETE, 0);
		break;
		case eCTRL_TAKE_ACTIVE_NEXT :
			Main_OnCommand(ID_MAIN_TAKE_ACTIVE_NEXT, 0);
		break;
		case eCTRL_TAKE_ACTIVE_PREVIOUS :
			Main_OnCommand(ID_MAIN_TAKE_ACTIVE_PREVIOUS, 0);
		break;
		case eCTRL_TAKE_REVERSE :
			Main_OnCommand(ID_MAIN_TAKE_REVERSE, 0);
		break;

		case eCTRL_ITEM_MUTE :
			Main_OnCommand(ID_MAIN_ITEM_TOGGLE_MUTE, 0);
		break;
		case eCTRL_ITEM_LOOP_SOURCE :
			Main_OnCommand(ID_MAIN_ITEM_LOOP_SOURCE, 0);
		break;
		case eCTRL_ITEM_LOOP_SECTION :
			Main_OnCommand(ID_MAIN_ITEM_LOOP_SECTION, 0);
		break;
		case eCTRL_ITEM_PROPERTIES :
			Main_OnCommand(ID_MAIN_ITEM_PROPERTIES, 0);
		break;

//! \todo Replace all simple eCTRL/Main_OnCommand() actions with map<enum,define> system
		case eCTRL_CURSEL_VIEW_ENVELOPES :
			Main_OnCommand(ID_MAIN_TRACK_VIEW_ENVELOPES, 0);
		break;
		case eCTRL_CURSEL_VIEW_FX :
			Main_OnCommand(ID_MAIN_TRACK_VIEW_FX, 0);
		break;
		case eCTRL_CURSEL_VIEW_IO :
			Main_OnCommand(ID_MAIN_TRACK_VIEW_IO, 0);
		break;
		case eCTRL_CURSEL_VIEW_GROUPING :
			Main_OnCommand(ID_MAIN_TRACK_VIEW_GROUPING, 0);
		break;

		case eCTRL_SHOW_ALL_ACTIVE_ENVELOPES :
			Main_OnCommand(ID_MAIN_ENVELOPE_TOGGLE_SHOW_ALL_ACTIVE, 0);
		break;
		case eCTRL_ENVELOPE_BYPASS :
			Main_OnCommand(ID_MAIN_ENVELOPE_TOGGLE_BYPASS, 0);
		break;
		case eCTRL_ENVELOPE_RECARM :
			Main_OnCommand(ID_MAIN_ENVELOPE_TOGGLE_RECARM, 0);
		break;
		case eCTRL_ENVELOPE_INSERT_POINT :
			Main_OnCommand(ID_MAIN_ENVELOPE_INSERT_POINT, 0);
		break;

		case eCTRL_ENVELOPE_VISIBLE_VOLUME :
			Main_OnCommand(ID_MAIN_ENVELOPE_TOGGLE_VISIBLE_VOLUME, 0);
		break;
		case eCTRL_ENVELOPE_VISIBLE_PAN :
			Main_OnCommand(ID_MAIN_ENVELOPE_TOGGLE_VISIBLE_PAN, 0);
		break;
		case eCTRL_ENVELOPE_VISIBLE_PREFXVOL :
			Main_OnCommand(ID_MAIN_ENVELOPE_TOGGLE_VISIBLE_PREFXVOL, 0);
		break;
		case eCTRL_ENVELOPE_VISIBLE_PREFXPAN :
			Main_OnCommand(ID_MAIN_ENVELOPE_TOGGLE_VISIBLE_PREFXPAN, 0);
		break;

		case eCTRL_ENVELOPE_CLEAR :
			Main_OnCommand(ID_MAIN_ENVELOPE_CLEAR, 0);
		break;
		case eCTRL_ENVELOPE_DELETE_POINTS :
			Main_OnCommand(ID_MAIN_ENVELOPE_DELETE_POINTS, 0);
		break;
		case eCTRL_ENVELOPE_TOGGLE_DISPLAY :
			Main_OnCommand(ID_MAIN_ENVELOPE_TOGGLE_DISPLAY, 0);
		break;
		case eCTRL_ENVELOPE_ALLVISIBLE_LANE :
			Main_OnCommand(ID_MAIN_ENVELOPE_TOGGLE_ALLVISIBLE_LANE, 0);
		break;

		case eCTRL_UNDO_HISTORY	:
			Main_OnCommand(ID_MAIN_VIEW_SHOW_UNDO_HISTORY, 0);
		break;

		case eCTRL_ITEM_PITCH_OCTAVE_WHEEL :
			if(value<0.5)
				Main_OnCommand(ID_MAIN_ITEM_PITCH_OCTAVE_DOWN, 0);
			else
				Main_OnCommand(ID_MAIN_ITEM_PITCH_OCTAVE_UP, 0);
		break;

		case eCTRL_ITEM_PITCH_ONECENT_WHEEL :
			if(value<0.5)
				Main_OnCommand(ID_MAIN_ITEM_PITCH_ONECENT_DOWN, 0);
			else
				Main_OnCommand(ID_MAIN_ITEM_PITCH_ONECENT_UP, 0);
		break;

		case eCTRL_ITEM_PITCH_SEMITONE_WHEEL :
			if(value<0.5)
				Main_OnCommand(ID_MAIN_ITEM_PITCH_SEMITONE_DOWN, 0);
			else
				Main_OnCommand(ID_MAIN_ITEM_PITCH_SEMITONE_UP, 0);
		break;

		case eCTRL_TAKESELECT_WHEEL :
			if(value<0.5)
				Main_OnCommand(ID_MAIN_TAKE_ACTIVE_PREVIOUS, 0);
			else
				Main_OnCommand(ID_MAIN_TAKE_ACTIVE_NEXT, 0);
		break;

		case eCTRL_ITEMSELECT_WHEEL :
			if(value<0.5)
				Main_OnCommand(ID_MAIN_ITEM_SELECT_PREVITEM, 0);
			else
				Main_OnCommand(ID_MAIN_ITEM_SELECT_NEXTITEM, 0);
		break;

		case eCTRL_MOVE_CURSOR_GRID_WHEEL :
			if(value<0.5)
				Main_OnCommand(ID_MAIN_MOVE_CURSOR_GRID_LEFT, 0);
			else
				Main_OnCommand(ID_MAIN_MOVE_CURSOR_GRID_RIGHT, 0);
		break;

		case eCTRL_AUTOMATION_ACTIVE_FADER_POT :
		{
			controller->setValue(value);
			int midi = Val2Midi(value);
			KBD_OnMainActionEx(ID_MAIN_AUTOMATION_ACTIVE_FADER_CCONLY, midi, -1, 0, NULL, NULL);
		}
		break;

		case eCTRL_AUTOMATION_ACTIVE_FADER_WHEEL :
			if(value<0.5)
				Main_OnCommand(ID_MAIN_AUTOMATION_ACTIVE_FADER_DEC, 0);
			else
				Main_OnCommand(ID_MAIN_AUTOMATION_ACTIVE_FADER_INC, 0);
		break;

		default :
		break;
	}

	if(controller->isEndless())
		_parent->addUpdateCtrl(controller);
}

int AutomapReaperClient::GetNumPresets() const
{
//	return 0;
	return _device->getNbPresets();
}

int AutomapReaperClient::GetCurrentPreset(char name[AUTOMAP_NAME_LENGTH+1]) const
{
//	return 0;
	return _device->getCurrentPreset(name);
}

void AutomapReaperClient::SetCurrentPreset(int n)
{
	if(_parent && (n<_device->getNbPresets()))
		_parent->resetMapping(n);
}

bool AutomapReaperClient::SupportsTransport(AutomapTransport type) const
{
	return true;
}

void AutomapReaperClient::OnTransport(AutomapTransport type, float param)
{
	switch(type)
	{
//! \todo Call CSurf_SetPlayState() (already done by CSurf_OnPlay() but "not to be relied upon")
		case TRANSPORT_PLAY : 
			if(param>=0.5)
				CSurf_OnPlay();
		break;
		case TRANSPORT_STOP : 
			if(param>=0.5)
				CSurf_OnStop();
		break;
		case TRANSPORT_RECORD : 
			if(param>=0.5)
				CSurf_OnRecord();
		break;
		case TRANSPORT_FORWARD : 
			setFfwRew(1, _parent->getFfwRwdMode(), (param>=0.5));
		break;
		case TRANSPORT_REWIND : 
			setFfwRew(-1, _parent->getFfwRwdMode(), (param>=0.5));
		break;
		case TRANSPORT_LOOP : 
			GetSetRepeat(1*(param>=0.5));
			//GetSetRepeat(2);
		break;

		case TRANSPORT_TEMPO :
		break;

		default :
		break;
	}
}

bool AutomapReaperClient::getTrackSendParamName(MediaTrack *tr, ReaperControlType controlType, int slot, char* paramName)
{
	int trackSendOffset = 0;
	int trackSendId = 0;
	int sendType = 0;
	char buffer[BUFFER_SIZE] = "???";

	switch(controlType)
	{
		case eCTRL_TRACKSEND_VOL : case eCTRL_TRACKSEND_PAN :
		case eCTRL_TRACKSEND_MUTE : case eCTRL_TRACKSEND_MODE :
		case eCTRL_TRACKSEND_MUTE_POT : case eCTRL_TRACKSEND_MODE_POT :
			trackSendOffset = controlType - eCTRL_TRACKSEND_VOL;
			trackSendId = _parent->getSelTrackSend() + trackSendOffset/6;
			sendType = 0;
			strcpy(buffer, "Send");
		break;

		case eCTRL_TRACKHARDSEND_VOL : case eCTRL_TRACKHARDSEND_PAN :
		case eCTRL_TRACKHARDSEND_MUTE : case eCTRL_TRACKHARDSEND_MODE :
		case eCTRL_TRACKHARDSEND_MUTE_POT : case eCTRL_TRACKHARDSEND_MODE_POT :
			trackSendOffset = controlType - eCTRL_TRACKHARDSEND_VOL;
			trackSendId = _parent->getSelTrackHardSend() + trackSendOffset/6;
			sendType = 1;
			strcpy(buffer, "HSend");
		break;
		default:
			return false;
		break;
	}

	trackSendId += slot;

	if(tr)
	{
		void* tmp = GetSetTrackSendInfo(tr, sendType, trackSendId, "D_VOL", NULL);
//		MediaTrack *dest = (MediaTrack*)(GetSetTrackSendInfo(tr, 0, trackSendId, "P_DESTTRACK", NULL));
		if(tmp)
		{
			switch(trackSendOffset%6)
			{
				case 0: AutomapReaperClient::sprintfCrop(paramName, "%s%d Vol", buffer, trackSendId+1); break;
				case 1: AutomapReaperClient::sprintfCrop(paramName, "%s%d Pan", buffer, trackSendId+1); break;
				case 2: case 4: AutomapReaperClient::sprintfCrop(paramName, "%s%d Mute", buffer, trackSendId+1); break;
				case 3: case 5: AutomapReaperClient::sprintfCrop(paramName, "%s%d Mode", buffer, trackSendId+1); break;
				default:
				break;
			}
		}
		else
			return false;
	}
	else
		return false;

	return true;
}

double AutomapReaperClient::getTrackSendValue(MediaTrack *tr, ReaperControlType controlType, int slot, char* valueTextOut)
{
	double value = 0.0;
	if(valueTextOut)
		AutomapReaperClient::strcpyCrop(valueTextOut, "???");

	int trackSendOffset = 0;
	int trackSendId = 0;
	int sendType = 0;

	switch(controlType)
	{
			case eCTRL_TRACKSEND_VOL : case eCTRL_TRACKSEND_PAN : case eCTRL_TRACKSEND_MUTE : case eCTRL_TRACKSEND_MODE :
			case eCTRL_TRACKSEND_MUTE_POT :
			case eCTRL_TRACKSEND_MODE_POT :
				trackSendOffset = controlType - eCTRL_TRACKSEND_VOL;
				trackSendId = _parent->getSelTrackSend() + trackSendOffset/6;
				sendType = 0;
			break;
			case eCTRL_TRACKHARDSEND_VOL : case eCTRL_TRACKHARDSEND_PAN : case eCTRL_TRACKHARDSEND_MUTE : case eCTRL_TRACKHARDSEND_MODE :
			case eCTRL_TRACKHARDSEND_MUTE_POT :
			case eCTRL_TRACKHARDSEND_MODE_POT :
				trackSendOffset = controlType - eCTRL_TRACKHARDSEND_VOL;
				trackSendId = _parent->getSelTrackHardSend() + trackSendOffset/6;
				sendType = 1;
			break;
			default:
				return value;
			break;
	}

	trackSendId += slot;

	if( tr && (trackSendId>=0) )
	{
		switch(trackSendOffset%6)
		{
			case 0 :
			{
				void* tmp = GetSetTrackSendInfo(tr, sendType, trackSendId, "D_VOL", NULL);
				if(tmp)
				{
					value = *(double*)(tmp);
					value = VAL2DB(value);
					value = DB2SLIDER(value)/1000.0;
					if(valueTextOut)
						_device->getVolValueStr(value, valueTextOut);
				}
			}
			break;

			case 1 :
			{
				void* tmp = GetSetTrackSendInfo(tr, sendType, trackSendId, "D_PAN", NULL);
				if(tmp)
				{
					value = *(double*)(tmp);
					value = value/2.0 + 0.5;
					if(valueTextOut)
						_device->getPanValueStr(value, valueTextOut);
				}
			}
			break;

			case 2 : case 4 :
			{
				void* tmp = GetSetTrackSendInfo(tr, sendType, trackSendId, "B_MUTE", NULL);
				if(tmp)
				{
					bool mute = *(bool*)(tmp);
					switch(sendType)
					{
						case 0	:	_parent->setSelTrackSendMute(mute);		break;
						case 1	:	_parent->setSelTrackHardSendMute(mute);	break;
						default	:	break;
					}
					value = 1.0 * mute;
					if(valueTextOut)
						_device->getButtonValueStr(value, valueTextOut);
				}
			}
			break;

			case 3 : case 5 :
			{
				void* tmp = GetSetTrackSendInfo(tr, sendType, trackSendId, "I_SENDMODE", NULL);
				if(tmp)
				{
					int mode = *(int*)(tmp);
					switch(sendType)
					{
						case 0	:	_parent->setSelTrackSendMode(mode);		break;
						case 1	:	_parent->setSelTrackHardSendMode(mode);	break;
						default	:	break;
					}

					if( (controlType==eCTRL_TRACKSEND_MODE_POT) || (controlType==eCTRL_TRACKHARDSEND_MODE_POT) )
						value = sendMode2PotVal(mode);
					else
						value = 1.0;
					if(valueTextOut)
						_device->getSendModeValueStr(mode, valueTextOut);
				}
			}
			break;

			default:
			break;
		}
	}

	return value;
}

double AutomapReaperClient::getFxParamValue(MediaTrack *tr, ReaperControlType controlType, int slot, char* valueTextOut)
{
	double value = 0.0;
	int fxParamOffset = controlType - eCTRL_FXPARAM;
	int fxParam = _parent->getSelFxParam()+fxParamOffset;
	fxParam += slot;

	if( tr && (fxParam>=0) )
	{
		double minval = 0.0;
		double maxval = 1.0;
		_parent->setFxParamMinMax(fxParam, minval, maxval);
		double realvalue = TrackFX_GetParam(tr, _parent->getSelFx(), fxParam, &minval, &maxval);
		value = realvalue;
		if(maxval>minval)
		{
			value = (realvalue-minval)/(maxval-minval);
			_parent->setFxParamMinMax(fxParam, minval, maxval);
		}

		if(valueTextOut)
		{
			//! \note TrackFX_FormatParamValue() only works if plugin implements Cockos VST extensions
			//! \note TrackFX_GetParam() min/max only works if plugin implements Cockos VST extensions
			char buffer[BUFFER_SIZE];
			bool res = TrackFX_FormatParamValue(tr, _parent->getSelFx(), fxParam, realvalue, buffer, AUTOMAP_NAME_LENGTH);
			if(res)
				AutomapReaperClient::strcpyCrop(valueTextOut, buffer);
			else
				AutomapReaperClient::sprintfCrop(valueTextOut, "%.0lf", 127.0*value);
		}
	}

	return value;
}

void AutomapReaperClient::setTrackSendValue(MediaTrack *tr, ReaperControlType controlType, int slot, double value)
{
	int trackSendOffset = 0;
	int trackSendId = 0;
	int sendType = 0;

	switch(controlType)
	{
			case eCTRL_TRACKSEND_VOL : case eCTRL_TRACKSEND_PAN : case eCTRL_TRACKSEND_MUTE : case eCTRL_TRACKSEND_MODE :
			case eCTRL_TRACKSEND_MUTE_POT :
			case eCTRL_TRACKSEND_MODE_POT :
				trackSendOffset = controlType - eCTRL_TRACKSEND_VOL;
				trackSendId = _parent->getSelTrackSend() + trackSendOffset/6;
				sendType = 0;
			break;
			case eCTRL_TRACKHARDSEND_VOL : case eCTRL_TRACKHARDSEND_PAN : case eCTRL_TRACKHARDSEND_MUTE : case eCTRL_TRACKHARDSEND_MODE :
			case eCTRL_TRACKHARDSEND_MUTE_POT :
			case eCTRL_TRACKHARDSEND_MODE_POT :
				trackSendOffset = controlType - eCTRL_TRACKHARDSEND_VOL;
				trackSendId = _parent->getSelTrackHardSend() + trackSendOffset/6;
				sendType = 1;
			break;
			default:
				return;
			break;
	}

	trackSendId += slot;

	if(tr && (trackSendId>=0))
	{
		switch(trackSendOffset%6)
		{
			case 0 :
			{
				value = SLIDER2DB(1000.0*value);
				value = DB2VAL(value);
				GetSetTrackSendInfo(tr, sendType, trackSendId, "D_VOL", (double*)(&value)); break;
			}
			break;

			case 1 :
			{
				value = 2.0*value - 1.0;
				GetSetTrackSendInfo(tr, sendType, trackSendId, "D_PAN", (double*)(&value)); break;
			}
			break;

			case 2 : case 4 :
			{
				bool mute = (value>=0.5);
				switch(sendType)
				{
					case 0	:	_parent->setSelTrackSendMute(mute);			break;
					case 1	:	_parent->setSelTrackHardSendMute(mute);		break;
					default	:	break;
				}
				GetSetTrackSendInfo(tr, sendType, trackSendId, "B_MUTE", (bool*)(&mute));
			}
			break;

			case 3 : case 5 :
			{
				int mode = 0;
				switch(sendType)
				{
					case 0	:	mode = _parent->getSelTrackSendMode();			break;
					case 1	:	mode = _parent->getSelTrackHardSendMode();		break;
					default	:	break;
				}

				if( (controlType==eCTRL_TRACKSEND_MODE_POT) || (controlType==eCTRL_TRACKHARDSEND_MODE_POT) )
					mode = potVal2SendMode(value);
				else
				{
					mode++;
					mode = mode%4;
					if(mode==2)
						mode++;
				}

				switch(sendType)
				{
					case 0	:	_parent->setSelTrackSendMode(mode);			break;
					case 1	:	_parent->setSelTrackHardSendMode(mode);		break;
					default	:	break;
				}
				//! \bug Reaper: GetSetTrackSendInfo("I_SENDMODE") does not refresh IO window
				GetSetTrackSendInfo(tr, sendType, trackSendId, "I_SENDMODE", (int*)(&mode)); break;
			}

			default:
			break;
		}
	}
}

void AutomapReaperClient::setFxParamValue(MediaTrack *tr, ReaperControlType controlType, int slot, double value)
{
	int fxParamOffset = controlType - eCTRL_FXPARAM;
	int fxParam = _parent->getSelFxParam()+fxParamOffset;
	fxParam += slot;

	if(tr && (fxParam>=0))
	{
		double minval = 0.0;
		double maxval = 1.0;
		if(_parent->hasFxParamMinMax(fxParam))
			_parent->getFxParamMinMax(fxParam, &minval, &maxval);

		double realvalue = (maxval-minval)*value + minval;
		TrackFX_SetParam(tr, _parent->getSelFx(), fxParam, realvalue);
	}
}

void AutomapReaperClient::setFfwRew(int direction, int mode, bool active)
{
	//if(dynamic_cast<AutomapiPhone*>(_device) != NULL)
	//{
	//	//! \todo Make useless
	//	if(mode == AutomapCSurf::Parameters::eFFWRWD_STANDARD)
	//		mode = AutomapCSurf::Parameters::eFFWRWD_MOVECURSOR;
	//}

	switch(mode)
	{
		case AutomapCSurf::Parameters::eFFWRWD_STANDARD :
			AutomapCSurf::setDoFfw( active && (direction>=0) );
			AutomapCSurf::setDoRew( active && (direction<0));
			//CSurf_OnFwd(_parent->_parameters.seekplay);
		break;
		case AutomapCSurf::Parameters::eFFWRWD_MOVECURSOR :
			if(active)
				MoveEditCursor(direction * AutomapCSurf::getFfwRwdRange(), false);
		break;
		case AutomapCSurf::Parameters::eFFWRWD_STARTEND :
			if(active)
			{
				if(direction<0)
					CSurf_GoStart();
				else
					CSurf_GoEnd();
			}
		break;
		case AutomapCSurf::Parameters::eFFWRWD_NEXTPREVMARKER :
			if(active)
			{
				if(direction<0)
					Main_OnCommand(ID_MARKER_PREV, 0);
				else
					Main_OnCommand(ID_MARKER_NEXT, 0);
			}
		break;
		default:
		break;
	}
}

double AutomapReaperClient::vol2Value(double volume, char* valueTextOut)
{
	double value = VAL2DB(volume);
	value = DB2SLIDER(value)/1000.0;
	if(valueTextOut)
		_device->getVolValueStr(value, valueTextOut);
		//mkvolstr(valueTextOut, volume);
	return value;
}

double AutomapReaperClient::value2Vol(double value)
{
	value = SLIDER2DB(1000.0*value);
	value = DB2VAL(value);
	return value;
}

double AutomapReaperClient::pan2Value(double pan, char* valueTextOut)
{
	double value = pan/2.0 + 0.5;
	if(valueTextOut)
		_device->getPanValueStr(value, valueTextOut);
		//mkpanstr(valueTextOut, pan);
	return value;
}

double AutomapReaperClient::value2Pan(double value)
{
	value = 2.0*value - 1.0;
	return value;
}

double AutomapReaperClient::boolStatus2Value(bool status, char* valueTextOut)
{
	double value = 1.0*status;
	if(valueTextOut != NULL)
		_device->getButtonValueStr(value, valueTextOut);
	return value;
}

double AutomapReaperClient::sendMode2PotVal(int mode)
{
	if(mode > 3)
		mode = 2;		// will put pre-fader to 2 for convenience
	return (double)(mode)/3.0;
}

int AutomapReaperClient::potVal2SendMode(double value)
{
	int mode = (int)(3.0*value);
	if(mode == 2)
		mode = 3;
	return mode;
}

AutomapSLRawMidiClient::MidiEvent::MidiEvent()
{
	evt.frame_offset = -1;
	evt.size = 0;
}

bool AutomapSLRawMidiClient::UniversalDeviceInquiryReply::isRemoteSL()
{
	char slMan[3]		= {0x00, 0x20, 0x29};
	char slFamily[2]	= {0x03, 0x00};
	char slMember[2]	= {0x03, 0x00};

	if( !memcmp(man, slMan, sizeof(slMan)) && !memcmp(family, slFamily, sizeof(slFamily)) && !memcmp(member, slMember, sizeof(slMember)) )
		return true;
	return false;
}

AutomapSLRawMidiClient::AutomapSLRawMidiClient()
{
}

AutomapSLRawMidiClient::AutomapSLRawMidiClient(int midiPort, int *errStats)
: _midiout(NULL)
{
	_midiout = midiPort >= 0 ? CreateThreadedMIDIOutput(CreateMIDIOutput(midiPort,false,NULL)) : NULL;

	if(errStats)
	{
		if (midiPort >=0  && !_midiout) *errStats|=2;
	}
}

AutomapSLRawMidiClient::~AutomapSLRawMidiClient()
{
	if (_midiout)
	{
		delete _midiout;
		_midiout = NULL;
	}
}

void AutomapSLRawMidiClient::appendMidiMsg(MidiEvent &msg, int nbBytes, ...)
{
	va_list valArguments;
	va_start(valArguments, nbBytes);
	for(int i=0; i<nbBytes; i++)
		msg.evt.midi_message[msg.evt.size++] = va_arg(valArguments,unsigned char);
	va_end(valArguments);
}

bool AutomapSLRawMidiClient::isValidInPort(int portId)
{
	char buf[BUFFER_SIZE];
	return GetMIDIInputName(portId,buf,sizeof(buf));
}

bool AutomapSLRawMidiClient::isValidOutPort(int portId)
{
	char buf[BUFFER_SIZE];
	return GetMIDIOutputName(portId,buf,sizeof(buf));
}

void AutomapSLRawMidiClient::sendUniversalDeviceInquiry(midi_Output* midiout)
{
	if(midiout)
	{
		unsigned char pid = 0x02;
		MidiEvent msg;
		appendMidiMsg(msg, 7, 0xF0, 0x7E, 0x7F, 0x06, 0x01, 0xF7, 0x00);
		midiout->SendMsg(&msg.evt, -1);
	}
}

bool AutomapSLRawMidiClient::detectDeviceMidiPorts(int &midiPortIn, int &midiPortOut)
{
	if(AutomapCSurf::_automapRawMidiClient)
	{
		delete AutomapCSurf::_automapRawMidiClient;
		AutomapCSurf::_automapRawMidiClient = NULL;
	}

	bool res = false;
	midiPortIn = -1;
	midiPortOut = -1;
	int numInputs = GetNumMIDIInputs();
	int numOutputs = GetNumMIDIOutputs();

	map<int, midi_Input*> inputs;
	for(int inId=0; inId<numInputs; inId++)
	{
			if(!isValidInPort(inId))
				continue;
			midi_Input* midiin = inId >= 0 ? CreateMIDIInput(inId) : NULL;
			if(midiin)
			{
				inputs[inId] = midiin;
				midiin->start();
			}
	}

	for(int outId=0; (outId<numOutputs) && !res; outId++)
	{
		if(!isValidOutPort(outId))
			continue;
		midi_Output* midiout = outId >= 0 ? CreateThreadedMIDIOutput(CreateMIDIOutput(outId,false,NULL)) : NULL;
		if(!midiout)
			continue;

		sendUniversalDeviceInquiry(midiout);
		Sleep(200);
		//! \note You have to wait for MIDI-In to kick back !!!

		for(map<int, midi_Input*>::iterator it = inputs.begin(); (it != inputs.end()) && !res; it++)
		{
			//it->second->SwapBufs(timeGetTime());
			it->second->SwapBufsPrecise(timeGetTime(), 1000.0*timeGetTime());
			MIDI_eventlist *list = it->second->GetReadBuf();
			MIDI_event_t *evt;
			int l=0;
			while(evt=list->EnumItems(&l))
			{
				UniversalDeviceInquiryReply udi;
				memcpy(&udi,evt->midi_message,17);
				if(udi.isRemoteSL())
				{
					midiPortIn = it->first;
					midiPortOut = outId;
					res = true;
				}
				l++;
			}
		}

		delete midiout;
		midiout = NULL;
	}

	for(map<int, midi_Input*>::iterator it = inputs.begin(); it != inputs.end(); it++)
	{
		if(it->second)
		{
			delete it->second;
			it->second = NULL;
		}
	}
	inputs.clear();

	return res;
}

void AutomapSLRawMidiClient::clearPanel(Panel panel)
{
	if(_midiout)
	{
		unsigned char pid = 0x02;
		MidiEvent msg;
		appendMidiMsg(msg, 15, 0xF0, 0x00, 0x20, 0x29, 0x03, 0x03, 0x20, 0x02, pid, 0x00, 0x02, 0x02, (unsigned char)(panel), 0xF7, 0x00);
		_midiout->SendMsg(&msg.evt, -1);
	}
}

void AutomapSLRawMidiClient::sendText(Row row, string text, unsigned char pos)
{
	if(_midiout)
	{
		unsigned char pid = 0x02;
		MidiEvent msg;
		appendMidiMsg(msg, 15, 0xF0, 0x00, 0x20, 0x29, 0x03, 0x03, 0x20, 0x02, pid, 0x00, 0x02, 0x01, pos, (unsigned char)(row), 0x04);
		for(unsigned int i=0; (i<text.size()) && (i<AUTOMAPSLDISPLAY_WIDTH-(unsigned int)(pos)); i++)
			appendMidiMsg(msg, 1, (unsigned char)(text[i]));
		appendMidiMsg(msg, 2, 0xF7, 0x00);
		_midiout->SendMsg(&msg.evt, -1);
	}
}

AutomapCSurf::Parameters::Parameters()
{
	deviceType = AutomapDevice::eREMOTESL;
	hasGlobalPages = true;
	mcpMode = false;
	zoomFactor = 1.0;
	ffwRwdMode = eFFWRWD_STANDARD;
	ffwRwdRange = 10.0;
	seekplay = 128;
	midiOut = -1;
	fnMidiChannel = 15;
	speedDialFactor = 128;
}

AutomapCSurf::Parameters AutomapCSurf::_parameters;
AutomapSLRawMidiClient* AutomapCSurf::_automapRawMidiClient = NULL;
bool AutomapCSurf::_doFfw = false;
bool AutomapCSurf::_doRew = false;

AutomapCSurf::AutomapCSurf(int *errStats)
: _selTrack(1), _selFx(0), _selFxParam(0), _selTcpFxParam(0), _selTrackSend(0), _selTrackHardSend(0), _nbTracksMax(0)
, _selTrackSendMode(0), _selTrackSendMute(false), _selTrackHardSendMode(0), _selTrackHardSendMute(false)
, _midiViewMode(0), _midiGrid(-1)
, _mainGrid(-1)
, _defaultEnvShape(-1), _selectedEnvShape(-1)
, _automapClient(NULL)
{
	_updateCtrls.clear();
	_fxParamsMinMax.clear();
	strcpy(_configString, "");

	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	_automapClient = new AutomapReaperClient(this);

	if(!AutomapCSurf::_automapRawMidiClient)
		AutomapCSurf::_automapRawMidiClient = new AutomapSLRawMidiClient(_parameters.midiOut, errStats);

	//! \todo MsgBox if reg error
	AutomapRegisterer::registerActions();
}

AutomapCSurf::~AutomapCSurf()
{
	CloseNoReset();
}

const char* AutomapCSurf::GetTypeString()
{
	return "Automap";
}

const char* AutomapCSurf::GetDescString()
{
	return "Novation Automap";
}

const char* AutomapCSurf::GetConfigString()
{
	parametersToConfigString(_parameters, _configString);
	return _configString;
}

void AutomapCSurf::CloseNoReset() 
{ 
	clearUpdateCtrls();

	AutomapRegisterer::unregisterActions();

	if(AutomapCSurf::_automapRawMidiClient)
	{
		delete AutomapCSurf::_automapRawMidiClient;
		AutomapCSurf::_automapRawMidiClient = NULL;
	}

	if(_automapClient)
	{
		delete _automapClient;
		_automapClient = NULL;
	}

	if(gMutex)
	{
		CloseHandle(gMutex);
		gMutex = NULL;
	}
}

void AutomapCSurf::updateControllers()
{
	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		for(set<AutomapController*>::iterator it = _updateCtrls.begin(); it != _updateCtrls.end(); it++)
			(*it)->refreshClient(_automapClient);
		_updateCtrls.clear();

		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}
}

void AutomapCSurf::Run()
{
	if(getDoFfw())
		CSurf_OnFwd(_parameters.seekplay);
	if(getDoRew())
		CSurf_OnRew(_parameters.seekplay);

	updateControllers();

	if(_automapClient)
		_automapClient->GetConnection()->Update();
}

void AutomapCSurf::SetTrackListChange()
{
	resetMapping(-1);
}

void AutomapCSurf::resetMapping(int preset)
{
	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		_updateCtrls.clear();
		_doFfw = false;
		_doRew = false;

		char buffer[BUFFER_SIZE];
		EnumProjects(-1, buffer, BUFFER_SIZE);
		string projectName(buffer);

		if(!projectName.empty())
			_automapClient->_instanceName = projectName.substr(projectName.find_last_of("/\\")+1);
		else
			_automapClient->_instanceName = "???";
	
		int nbTracks = CSurf_NumTracks(_parameters.mcpMode);
		// Dirty fix for Automap Server crash: make sure the number of pages never decreases
		if(_nbTracksMax<nbTracks)
			_nbTracksMax = nbTracks;

//_automapClient->_nbParams = 0;
//_automapClient->GetConnection()->ParamCountChanged();

		//_automapClient->_device->initControllers(_nbTracksMax);
		_automapClient->_device->initControllers(_nbTracksMax, preset);
		_automapClient->_nbParams = _automapClient->_device->getNbParams();
		_automapClient->_device->initControllers(nbTracks);

		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		}
	}

	_automapClient->GetConnection()->InstanceNameChanged();
	_automapClient->GetConnection()->ParamCountChanged();
}

void AutomapCSurf::SetSurfaceVolume(MediaTrack *trackid, double volume) 
{
	int trackIdx = CSurf_TrackToID(trackid,getMcpMode());
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_VOL, trackIdx))
		ctrl->refreshClient(_automapClient);
	if(trackIdx == getSelTrack())
	{
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_VOL, -1))
			ctrl->refreshClient(_automapClient);
	}
}

void AutomapCSurf::SetSurfacePan(MediaTrack *trackid, double pan) 
{
	int trackIdx = CSurf_TrackToID(trackid,getMcpMode());
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_PAN, trackIdx))
		ctrl->refreshClient(_automapClient);
	if(trackIdx == getSelTrack())
	{
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_PAN, -1))
			ctrl->refreshClient(_automapClient);
	}
}

void AutomapCSurf::SetSurfaceMute(MediaTrack *trackid, bool mute) 
{
	int trackIdx = CSurf_TrackToID(trackid,getMcpMode());
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_MUTE, trackIdx))
		ctrl->refreshClient(_automapClient);
	if(trackIdx == getSelTrack())
	{
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_MUTE, -1))
			ctrl->refreshClient(_automapClient);
	}
}

void AutomapCSurf::SetSurfaceSelected(MediaTrack *trackid, bool selected) 
{
	int trackIdx = CSurf_TrackToID(trackid,getMcpMode());
	if(selected)
	{
		setSelTrack(trackIdx);

		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSELECT_WHEEL, -1))
			ctrl->refreshClient(_automapClient);

		int numTracks = CSurf_NumTracks(false);
		for(int track=0; track<numTracks; track++)
		{
			if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSELECT, track))
			ctrl->refreshClient(_automapClient);
		}

		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_SUBMENUSELECT_WHEEL, -1))
			ctrl->refreshClient(_automapClient, true);

		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_VOL, -1))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_PAN, -1))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_MUTE, -1))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_SOLO, -1))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_ARM, -1))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_FXBYPASS, -1))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_PHASE, -1))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_AUTOMATION_MODE, -1))
			ctrl->refreshClient(_automapClient, true);

		selectFx(trackid, 0);
		selectTrackSend(trackid, 0);
		selectHardTrackSend(trackid, 0);
	}
}

void AutomapCSurf::SetSurfaceSolo(MediaTrack *trackid, bool solo) 
{
	int trackIdx = CSurf_TrackToID(trackid,getMcpMode());
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_SOLO, trackIdx))
		ctrl->refreshClient(_automapClient);
	if(trackIdx == getSelTrack())
	{
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_SOLO, -1))
			ctrl->refreshClient(_automapClient);
	}
}

void AutomapCSurf::SetSurfaceRecArm(MediaTrack *trackid, bool recarm) 
{ 
	int trackIdx = CSurf_TrackToID(trackid,getMcpMode());
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_ARM, trackIdx))
		ctrl->refreshClient(_automapClient);
	if(trackIdx == getSelTrack())
	{
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_ARM, -1))
			ctrl->refreshClient(_automapClient);
	}
}

void AutomapCSurf::SetPlayState(bool play, bool pause, bool rec) 
{
	_automapClient->GetConnection()->SendTransport(TRANSPORT_RECORD, 1.0f*(rec && play && !pause));
	_automapClient->GetConnection()->SendTransport(TRANSPORT_PLAY, 1.0f*(play && !pause));

	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_PLAY, -1))
		ctrl->refreshClient(_automapClient);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_STOP, -1))
		ctrl->refreshClient(_automapClient);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_RECORD, -1))
		ctrl->refreshClient(_automapClient);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_RWD, -1))
		ctrl->refreshClient(_automapClient);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_FFW, -1))
		ctrl->refreshClient(_automapClient);
}

void AutomapCSurf::SetRepeatState(bool rep) 
{
	_automapClient->GetConnection()->SendTransport(TRANSPORT_LOOP, 1.0f*rep);

	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_LOOP, -1))
		ctrl->refreshClient(_automapClient);
}

void AutomapCSurf::SetTrackTitle(MediaTrack* trackid, const char *title)
{
	int trackIdx = CSurf_TrackToID(trackid,getMcpMode());

	if(trackIdx>0)
	{
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_VOL, trackIdx))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_PAN, trackIdx))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_MUTE, trackIdx))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_SOLO, trackIdx))
			ctrl->refreshClient(_automapClient, true);
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_ARM, trackIdx))
			ctrl->refreshClient(_automapClient, true);
	}

	if(trackIdx == getSelTrack())
	{
		if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSELECT_WHEEL, -1))
			ctrl->refreshClient(_automapClient);
	}
}

bool AutomapCSurf::GetTouchState(MediaTrack *trackid, int isPan) 
{
	return false;
}

void AutomapCSurf::SetAutoMode(int mode)
{
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_AUTOMATION_MODE, -1))
		ctrl->refreshClient(_automapClient);
}

void AutomapCSurf::ResetCachedVolPanStates() 
{
}

//! \todo Reaper: is this method useful?
void AutomapCSurf::OnTrackSelection(MediaTrack *trackid) 
{
/*
	int trackIdx = CSurf_TrackToID(trackid,getMcpMode());

if(trackIdx<0)
trackIdx = 0;


	setSelTrack(trackIdx);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSELECT_WHEEL, -1))
		ctrl->refreshClient(_automapClient, true);




////int numTracks = CSurf_NumTracks(getMcpMode());
//int numTracks = CSurf_NumTracks(false);
//for(int track=0; track<numTracks; track++)
//{
//if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSELECT, track))
//ctrl->refreshClient(_automapClient, true);
//}



	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_VOL, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_PAN, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_MUTE, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_SOLO, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_ARM, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_FXBYPASS, -1))
		ctrl->refreshClient(_automapClient, true);

	selectFx(trackid, 0);
	selectTrackSend(trackid, 0);
	selectHardTrackSend(trackid, 0);
*/
}

bool AutomapCSurf::IsKeyDown(int key) 
{
	return false; 
}

int AutomapCSurf::Extended(int call, void *parm1, void *parm2, void *parm3) 
{
	return 0;
}

#define CFLAG_GLOBALPAGES		(1<<0)
#define CFLAG_MCPMODE			(1<<1)

int AutomapCSurf::parametersToConfigFlags(AutomapCSurf::Parameters &parameters)
{
	int configFlags = 1*parameters.hasGlobalPages * CFLAG_GLOBALPAGES;
	configFlags += 1*parameters.mcpMode * CFLAG_MCPMODE;
	return configFlags;
}

void AutomapCSurf::configFlagsToParameters(int configFlags, AutomapCSurf::Parameters &parameters)
{
	parameters.hasGlobalPages = ( (configFlags&CFLAG_GLOBALPAGES) != 0 );
	parameters.mcpMode = ( (configFlags & CFLAG_MCPMODE) != 0 );
}

void AutomapCSurf::parametersToConfigString(Parameters &parameters, char* configString)
{
	int configFlags = AutomapCSurf::parametersToConfigFlags(parameters);
	sprintf(configString, "device=%d cflags=%d seekplay=%d zoomFactor=%.2lf ffwRwdMode=%d ffwRwdRange=%.2lf, midiOut=%d",
		parameters.deviceType, configFlags, parameters.seekplay, parameters.zoomFactor, parameters.ffwRwdMode, parameters.ffwRwdRange, parameters.midiOut);
}

bool AutomapCSurf::configStringToParameters(const char* configString, Parameters &parameters)
{
	int configFlags;
	int res = sscanf(configString, "device=%d cflags=%d seekplay=%d zoomFactor=%lf ffwRwdMode=%d ffwRwdRange=%lf, midiOut=%d",
					 &parameters.deviceType, &configFlags, &parameters.seekplay, &parameters.zoomFactor, &parameters.ffwRwdMode, &parameters.ffwRwdRange, &parameters.midiOut);
	if(res>0)
	{
		AutomapCSurf::configFlagsToParameters(configFlags, parameters);
		return true;
	}

	return false;
}

void AutomapCSurf::loadConfigFile()
{
	char filename[BUFFER_SIZE];
	strcpy(filename, GetExePath());
	strcat(filename, "\\Plugins\\reaper_csurf_automap.ini");

	FILE* file = fopen(filename, "r");
	if(file)
	{
		int tmpint;
		char tmpstr[BUFFER_SIZE];
		char buffer[BUFFER_SIZE];

		while(!feof(file))
		{
			fgets(buffer, BUFFER_SIZE, file);

			if(sscanf(buffer, "fnMidiChan=%d\n", &tmpint) == 1)
			{
				if(tmpint<0)
					tmpint = 0;
				if(tmpint>15)
					tmpint = 15;
				AutomapCSurf::setFnMidiChannel(tmpint);
			}

			else if(sscanf(buffer, "fnButAlias[%d]=%s", &tmpint, &tmpstr) == 2)
				AutomapCSurf::setFnButAlias(tmpint, tmpstr);

			else if(sscanf(buffer, "fnPotAlias[%d]=%s", &tmpint, &tmpstr) == 2)
				AutomapCSurf::setFnPotAlias(tmpint, tmpstr);

			else if(sscanf(buffer, "fnEncAlias[%d]=%s", &tmpint, &tmpstr) == 2)
				AutomapCSurf::setFnEncAlias(tmpint, tmpstr);
		}

		fclose(file);
	}
}

void AutomapCSurf::saveConfigFile()
{
	char filename[BUFFER_SIZE];
	strcpy(filename, GetExePath());
	strcat(filename, "\\Plugins\\reaper_csurf_automap.ini");

	FILE* file = fopen(filename, "w+");
	if(file)
	{
		fprintf(file, "fnMidiChan=%d\n", AutomapCSurf::getFnMidiChannel());

		if(!gMutex)
			gMutex = CreateMutex(NULL,FALSE,NULL);

		if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
		{
			for(map<int,string>::iterator it = _parameters.fnButAliases.begin(); it != _parameters.fnButAliases.end(); it++)
				fprintf(file, "fnButAlias[%d]=%s\n", it->first, it->second.c_str());
			for(map<int,string>::iterator it = _parameters.fnPotAliases.begin(); it != _parameters.fnPotAliases.end(); it++)
				fprintf(file, "fnPotAlias[%d]=%s\n", it->first, it->second.c_str());
			for(map<int,string>::iterator it = _parameters.fnEncAliases.begin(); it != _parameters.fnEncAliases.end(); it++)
				fprintf(file, "fnEncAlias[%d]=%s\n", it->first, it->second.c_str());

			if (!ReleaseMutex(gMutex)) 
			{
				// handle error
			} 
		}

		fclose(file);
	}
}

void AutomapCSurf::setDoubleThreadSafe(double *dest, double val)
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		*dest = val;
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}
}

double AutomapCSurf::getDoubleThreadSafe(double *value)
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	double tmp = 0.0;
	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		tmp = *value;
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	return tmp;
}

void AutomapCSurf::setIntThreadSafe(int *dest, int val)
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		*dest = val;
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}
}

int AutomapCSurf::getIntThreadSafe(int *value)
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	int tmp = 0;
	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		tmp = *value;
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	return tmp;
}

bool AutomapCSurf::hasFxParamMinMax(int fxParam)
{
	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		return (_fxParamsMinMax.count(fxParam) > 0);

		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	return false;
}

void AutomapCSurf::getFxParamMinMax(int fxParam, double *minval, double *maxval)
{
	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		*minval = _fxParamsMinMax[fxParam].first;
		*maxval = _fxParamsMinMax[fxParam].second;

		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}
}

void AutomapCSurf::setFxParamMinMax(int fxParam, double minval, double maxval)
{
	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		_fxParamsMinMax[fxParam] = make_pair(minval,maxval);

		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}
}

void AutomapCSurf::setDoFfw(bool status)
{
	setBoolThreadSafe(&_doFfw, status);
}

bool AutomapCSurf::getDoFfw()
{
	return getBoolThreadSafe(&_doFfw);
}

void AutomapCSurf::setDoRew(bool status)
{
	setBoolThreadSafe(&_doRew, status);
}

bool AutomapCSurf::getDoRew()
{
	return getBoolThreadSafe(&_doRew);
}

void AutomapCSurf::toggleDoFfw()
{
	bool status = getBoolThreadSafe(&_doFfw);
	setBoolThreadSafe(&_doFfw, !status);
}

void AutomapCSurf::toggleDoRew()
{
	bool status = getBoolThreadSafe(&_doRew);
	setBoolThreadSafe(&_doRew, !status);
}

void AutomapCSurf::addUpdateCtrl(AutomapController* controller)
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		_updateCtrls.insert(controller);

		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}
}

void AutomapCSurf::clearUpdateCtrls()
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		_updateCtrls.clear();

		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}
}

void AutomapCSurf::toggleMcpMode()
{
	bool mcpMode = getMcpMode();
	setMcpMode(!mcpMode);
}

void AutomapCSurf::setMcpMode(bool val)
{
	setBoolThreadSafe(&_parameters.mcpMode, val);
	TrackList_UpdateAllExternalSurfaces();
}

bool AutomapCSurf::getMcpMode()
{
	return getBoolThreadSafe(&_parameters.mcpMode);
}

void AutomapCSurf::setFfwRwdMode(int mode)
{
	setIntThreadSafe((int*)(&_parameters.ffwRwdMode), mode);
}

int AutomapCSurf::getFfwRwdMode()
{
	return getIntThreadSafe((int*)(&_parameters.ffwRwdMode));
}

void AutomapCSurf::toggleFfwRwdMode()
{
	setDoFfw(false);
	setDoRew(false);
	int mode = getFfwRwdMode();
	int nbModes = Parameters::eFFWRWD_NBMODES;
	mode = (mode+1)%nbModes;
	setFfwRwdMode(mode);
}

void AutomapCSurf::setZoomFactor(double factor)
{
	setDoubleThreadSafe(&_parameters.zoomFactor, factor);
}

double AutomapCSurf::getZoomFactor()
{
	return getDoubleThreadSafe(&_parameters.zoomFactor);
}

void AutomapCSurf::increaseZoomFactor()
{
	double factor = getZoomFactor();
	factor *= 2.0;
	setZoomFactor(factor);
}

void AutomapCSurf::decreaseZoomFactor()
{
	double factor = getZoomFactor();
	factor /= 2.0;
	setZoomFactor(factor);
}

void AutomapCSurf::setSpeedDialFactor(int factor)
{
	setIntThreadSafe(&_parameters.speedDialFactor, factor);
}

int AutomapCSurf::getSpeedDialFactor()
{
	return getIntThreadSafe(&_parameters.speedDialFactor);
}

void AutomapCSurf::increaseSpeedDialFactor()
{
	int factor = getSpeedDialFactor();
	factor *= 2;
	setSpeedDialFactor(factor);
}

void AutomapCSurf::decreaseSpeedDialFactor()
{
	int factor = getSpeedDialFactor();
	if(factor>1)
		factor /= 2;
	setSpeedDialFactor(factor);
}

void AutomapCSurf::setFfwRwdRange(double factor)
{
	setDoubleThreadSafe(&_parameters.ffwRwdRange, factor);
}

double AutomapCSurf::getFfwRwdRange()
{
	return getDoubleThreadSafe(&_parameters.ffwRwdRange);
}

void AutomapCSurf::increaseFfwRwdRange()
{
	double factor = getFfwRwdRange();
	factor *= 2.0;
	setFfwRwdRange(factor);
}

void AutomapCSurf::decreaseFfwRwdRange()
{
	double factor = getFfwRwdRange();
	factor /= 2.0;
	setFfwRwdRange(factor);
}

int AutomapCSurf::getFnMidiChannel()
{
	return getIntThreadSafe(&_parameters.fnMidiChannel);
}

void AutomapCSurf::setFnMidiChannel(int channel)
{
	setIntThreadSafe(&_parameters.fnMidiChannel, channel);
}

const char* AutomapCSurf::getFnButAlias(int slot)
{
	const char* ptr = NULL;

	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		if(_parameters.fnButAliases.count(slot) != 0)
			ptr = _parameters.fnButAliases[slot].c_str();
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	return ptr;
}

void AutomapCSurf::setFnButAlias(int slot, const char* alias)
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		_parameters.fnButAliases[slot] = string(alias);
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	TrackList_UpdateAllExternalSurfaces();
}

const char* AutomapCSurf::getFnPotAlias(int slot)
{
	const char* ptr = NULL;

	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		if(_parameters.fnPotAliases.count(slot) != 0)
			ptr = _parameters.fnPotAliases[slot].c_str();
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	return ptr;
}

void AutomapCSurf::setFnPotAlias(int slot, const char* alias)
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		_parameters.fnPotAliases[slot] = string(alias);
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	TrackList_UpdateAllExternalSurfaces();
}

const char* AutomapCSurf::getFnEncAlias(int slot)
{
	const char* ptr = NULL;

	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		if(_parameters.fnEncAliases.count(slot) != 0)
			ptr = _parameters.fnEncAliases[slot].c_str();
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	return ptr;
}

void AutomapCSurf::setFnEncAlias(int slot, const char* alias)
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		_parameters.fnEncAliases[slot] = string(alias);
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	TrackList_UpdateAllExternalSurfaces();
}

void AutomapCSurf::resetFnAliases()
{
	if(!gMutex)
		gMutex = CreateMutex(NULL,FALSE,NULL);

	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		_parameters.fnButAliases.clear();
		_parameters.fnPotAliases.clear();
		_parameters.fnEncAliases.clear();
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}

	TrackList_UpdateAllExternalSurfaces();
}

void AutomapCSurf::selectTrack(int trackIndex)
{
	int numTracks = CSurf_NumTracks(getMcpMode());

	if(trackIndex<0)
		trackIndex = 0;
	else if(trackIndex>=numTracks)
		trackIndex = numTracks;

	MediaTrack *sel = CSurf_TrackFromID(trackIndex, getMcpMode());
	while( !sel && (trackIndex>0) && (trackIndex<=numTracks) )
	{
		sel = CSurf_TrackFromID(trackIndex, getMcpMode());
		trackIndex++;
	}

	if(sel)
	{
		clearSelectedTracks();
		CSurf_SetSurfaceSelected(sel, CSurf_OnSelectedChange(sel,-1),NULL);
	}
}

void AutomapCSurf::selectFx(MediaTrack* trackid, int fx)
{
	int nbFxs = 0;
	if(trackid)
		nbFxs = TrackFX_GetCount(trackid);
	if(fx<0)
		fx = 0;
	if(fx>=nbFxs)
		fx = nbFxs-1;

	setSelFx(fx);

	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_FXSELECT_WHEEL, -1))
		ctrl->refreshClient(_automapClient, true);

	_fxParamsMinMax.clear();
	selectFxParam(trackid, 0);
}

void AutomapCSurf::selectFxParam(MediaTrack* trackid, int param)
{
	int nbFxParams = 0;
	if(trackid)
		nbFxParams = TrackFX_GetNumParams(trackid, getSelFx());
	int maxOffset = nbFxParams - _automapClient->_device->getNbFxParamSlots();
	if(param>maxOffset)
		param = maxOffset;
	if(param<0)
		param = 0;
	if(param>=nbFxParams)
		param = nbFxParams-1;

	setSelFxParam(param);

	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_FXPARAMSELECT_WHEEL, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_FXPARAM, -1))
		ctrl->refreshClient(_automapClient, true);
}

void AutomapCSurf::selectTcpFxParam(int param)
{
	int nbTcpFxParams = 16;
	int maxOffset = nbTcpFxParams - _automapClient->_device->getNbTcpFxParamSlots();
	if(param>maxOffset)
		param = maxOffset;
	if(param<0)
		param = 0;
	if(param>=nbTcpFxParams)
		param = nbTcpFxParams-1;

	setSelTcpFxParam(param);

	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_TCPFXPARAMSELECT_WHEEL, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TCPFXPARAM, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TCPFXPARAM_ENC, -1))
		ctrl->refreshClient(_automapClient, true);
}

void AutomapCSurf::selectTrackSend(MediaTrack* trackid, int sendId)
{
	int nbTrackSends = 0;
	if(trackid)
	{
		//! \note I'll stop at 100 (unless you're a maniac)
		for(nbTrackSends=0; nbTrackSends<100; nbTrackSends++)
		{
			MediaTrack *tmp = (MediaTrack*)(GetSetTrackSendInfo(trackid, 0, nbTrackSends, "P_DESTTRACK", NULL));
			if(!tmp)
				break;
		}
	}

	int maxOffset = nbTrackSends - _automapClient->_device->getNbTrackSendSlots();
	if(sendId>maxOffset)
		sendId = maxOffset;
	if(sendId<0)
		sendId = 0;
	if(sendId>=nbTrackSends)
		sendId = nbTrackSends-1;

	setSelTrackSend(sendId);

	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_TRACKSENDSELECT_WHEEL, -1))
		ctrl->refreshClient(_automapClient, true);

	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSEND_VOL, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSEND_PAN, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSEND_MUTE, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSEND_MODE, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSEND_MUTE_POT, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKSEND_MODE_POT, -1))
		ctrl->refreshClient(_automapClient, true);
}

void AutomapCSurf::selectHardTrackSend(MediaTrack* trackid, int hardSendId)
{
	int nbTrackHardSends = 0;
	if(trackid)
	{
		//! \note I'll stop at 100 (unless you're a maniac)
		for(nbTrackHardSends=0; nbTrackHardSends<100; nbTrackHardSends++)
		{
			void* tmp = GetSetTrackSendInfo(trackid, 1, nbTrackHardSends, "D_VOL", NULL);
			if(!tmp)
				break;
		}
	}

	int maxOffset = nbTrackHardSends - _automapClient->_device->getNbTrackHardSendSlots();
	if(hardSendId>maxOffset)
		hardSendId = maxOffset;
	if(hardSendId<0)
		hardSendId = 0;
	if(hardSendId>=nbTrackHardSends)
		hardSendId = nbTrackHardSends-1;

	setSelTrackHardSend(hardSendId);

	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_CURSEL_TRACKHARDSENDSELECT_WHEEL, -1))
		ctrl->refreshClient(_automapClient, true);

	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKHARDSEND_VOL, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKHARDSEND_PAN, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKHARDSEND_MUTE, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKHARDSEND_MODE, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKHARDSEND_MUTE_POT, -1))
		ctrl->refreshClient(_automapClient, true);
	if(AutomapController* ctrl = _automapClient->_device->getController(eCTRL_TRACKHARDSEND_MODE_POT, -1))
		ctrl->refreshClient(_automapClient, true);
}

void AutomapCSurf::clearSelectedTracks()
{
	int numTracks = CSurf_NumTracks(false);
	for(int i=0; i<=numTracks; i++)
	{
		MediaTrack *tr = CSurf_TrackFromID(i, false);
	    if(tr)
			CSurf_SetSurfaceSelected(tr,CSurf_OnSelectedChange(tr,0), this);
//SetTrackSelected(tr, 0);
	}
}

void AutomapCSurf::selectMidiGrid(int grid)
{
	if(grid<0)
		grid = 0;
	if(grid>((int)(eGRID_LAST)-1))
		grid = (int)(eGRID_LAST)-1;

	setMidiGrid(grid);

	switch(grid)
	{
		case eGRID_4		:	MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE4, 0);		break;
		case eGRID_2		:	MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE2, 0);		break;
		case eGRID_1		:	MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1, 0);		break;
		case eGRID_1_2		:	MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_2, 0);		break;
		case eGRID_1_3		:	MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_3, 0);		break;
		case eGRID_1_4		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_2, 0);
			MidiEdit_OnCommand(ID_MIDI_GRID_DIVSIZE2, 0);
		break;
		case eGRID_1_4T		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_6, 0);
		break;
		case eGRID_1_8		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_16, 0);
			MidiEdit_OnCommand(ID_MIDI_GRID_MULTSIZE2, 0);
		break;
		case eGRID_1_8T		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_12, 0);
		break;
		case eGRID_1_16		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_16, 0);
		break;
		case eGRID_1_16T		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_24, 0);
		break;
		case eGRID_1_32		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_32, 0);
		break;
		case eGRID_1_32T		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_48, 0);
		break;
		case eGRID_1_64		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_32, 0);
			MidiEdit_OnCommand(ID_MIDI_GRID_DIVSIZE2, 0);
		break;
		case eGRID_1_64T		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_32, 0);
			MidiEdit_OnCommand(ID_MIDI_GRID_DIVSIZE3, 0);
		break;
		case eGRID_1_128		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_32, 0);
			MidiEdit_OnCommand(ID_MIDI_GRID_DIVSIZE2, 0);
			MidiEdit_OnCommand(ID_MIDI_GRID_DIVSIZE2, 0);
		break;
		case eGRID_1_128T		:
			MidiEdit_OnCommand(ID_MIDI_GRID_SETSIZE1_32, 0);
			MidiEdit_OnCommand(ID_MIDI_GRID_DIVSIZE3, 0);
			MidiEdit_OnCommand(ID_MIDI_GRID_DIVSIZE2, 0);
		break;
		default :
			return;
		break;
	}
}

void AutomapCSurf::selectMainGrid(int grid)
{
	if(grid<0)
		grid = 0;
	if(grid>((int)(eGRID_LAST)-1))
		grid = (int)(eGRID_LAST)-1;

	setMainGrid(grid);

	switch(grid)
	{
		case eGRID_4		:	Main_OnCommand(ID_MAIN_GRID_SET_4, 0);		break;
		case eGRID_2		:	Main_OnCommand(ID_MAIN_GRID_SET_2, 0);		break;
		case eGRID_1		:	Main_OnCommand(ID_MAIN_GRID_SET_1, 0);		break;
		case eGRID_1_2		:	Main_OnCommand(ID_MAIN_GRID_SET_1_2, 0);	break;
		case eGRID_1_3		:	Main_OnCommand(ID_MAIN_GRID_SET_1_3, 0);	break;
		case eGRID_1_4		:	Main_OnCommand(ID_MAIN_GRID_SET_1_4, 0);	break;
		case eGRID_1_4T		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_2, 0);
			Main_OnCommand(ID_MAIN_GRID_ADJUST_1_3, 0);
		break;
		case eGRID_1_8		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_8, 0);
		break;
		case eGRID_1_8T		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_4, 0);
			Main_OnCommand(ID_MAIN_GRID_ADJUST_1_3, 0);
		break;
		case eGRID_1_16		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_16, 0);
		break;
		case eGRID_1_16T		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_8, 0);
			Main_OnCommand(ID_MAIN_GRID_ADJUST_1_3, 0);
		break;
		case eGRID_1_32		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_32, 0);
		break;
		case eGRID_1_32T		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_16, 0);
			Main_OnCommand(ID_MAIN_GRID_ADJUST_1_3, 0);
		break;
		case eGRID_1_64		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_64, 0);
		break;
		case eGRID_1_64T		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_32, 0);
			Main_OnCommand(ID_MAIN_GRID_ADJUST_1_3, 0);
		break;
		case eGRID_1_128		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_64, 0);
			Main_OnCommand(ID_MAIN_GRID_ADJUST_1_2, 0);
		break;
		case eGRID_1_128T		:
			Main_OnCommand(ID_MAIN_GRID_SET_1_64, 0);
			Main_OnCommand(ID_MAIN_GRID_ADJUST_1_3, 0);
		break;
		default :
			return;
		break;
	}
}

void AutomapCSurf::selectMidiCcLane(int lane)
{
	switch(lane)
	{
		case 0	: MidiEdit_OnCommand(ID_MIDI_SETCCLANE_VELOCITY, 0);			break;
		case 1	: MidiEdit_OnCommand(ID_MIDI_SETCCLANE_PITCH, 0);				break;
		case 2	: MidiEdit_OnCommand(ID_MIDI_SETCCLANE_PROGRAM, 0);				break;
		case 3	: MidiEdit_OnCommand(ID_MIDI_SETCCLANE_CHANNELPRES, 0);			break;
		case 4	: MidiEdit_OnCommand(ID_MIDI_SETCCLANE_BANKPROGSELECT, 0);		break;
		case 5	: MidiEdit_OnCommand(ID_MIDI_SETCCLANE_TEXTEVENTS, 0);			break;
		case 6	: MidiEdit_OnCommand(ID_MIDI_SETCCLANE_SYSEX, 0);				break;
		default	:
		{
			lane -= 7;
			if(lane<0)
				lane = 0;
			if(lane>119)
				lane = 119;
			MidiEdit_OnCommand(ID_MIDI_SETCCLANE_000_BANKSELECT+lane, 0);
		}
		break;
	}
}

void AutomapCSurf::setBoolThreadSafe(bool *dest, bool val)
{
	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		*dest = val;
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}
}

bool AutomapCSurf::getBoolThreadSafe(bool *value)
{
	bool tmp = false;
	if(!WaitForSingleObject(gMutex, MUTEX_TIMEOUT))
	{
		tmp = *value;
		if (!ReleaseMutex(gMutex)) 
		{
			// handle error
		} 
	}
	return tmp;
}

void AutomapCSurf::setSelTrack(int val)
{
	setIntThreadSafe(&_selTrack, val);
}

int AutomapCSurf::getSelTrack()
{
	return getIntThreadSafe(&_selTrack);
}

void AutomapCSurf::setSelFx(int val)
{
	setIntThreadSafe(&_selFx, val);
}

int AutomapCSurf::getSelFx()
{
	return getIntThreadSafe(&_selFx);
}

void AutomapCSurf::setSelFxParam(int val)
{
	setIntThreadSafe(&_selFxParam, val);
}

int AutomapCSurf::getSelFxParam()
{
	return getIntThreadSafe(&_selFxParam);
}

void AutomapCSurf::setSelTcpFxParam(int val)
{
	setIntThreadSafe(&_selTcpFxParam, val);
}

int AutomapCSurf::getSelTcpFxParam()
{
	return getIntThreadSafe(&_selTcpFxParam);
}

void AutomapCSurf::setSelTrackSend(int val)
{
	setIntThreadSafe(&_selTrackSend, val);
}

int AutomapCSurf::getSelTrackSend()
{
	return getIntThreadSafe(&_selTrackSend);
}

void AutomapCSurf::setSelTrackHardSend(int val)
{
	setIntThreadSafe(&_selTrackHardSend, val);
}

int AutomapCSurf::getSelTrackHardSend()
{
	return getIntThreadSafe(&_selTrackHardSend);
}

void AutomapCSurf::setSelTrackSendMode(int val)
{
	setIntThreadSafe(&_selTrackSendMode, val);
}

int AutomapCSurf::getSelTrackSendMode()
{
	return getIntThreadSafe(&_selTrackSendMode);
}

void AutomapCSurf::setSelTrackSendMute(bool val)
{
	setBoolThreadSafe(&_selTrackSendMute, val);
}

bool AutomapCSurf::getSelTrackSendMute()
{
	return getBoolThreadSafe(&_selTrackSendMute);
}

void AutomapCSurf::setSelTrackHardSendMode(int val)
{
	setIntThreadSafe(&_selTrackHardSendMode, val);
}

int AutomapCSurf::getSelTrackHardSendMode()
{
	return getIntThreadSafe(&_selTrackHardSendMode);
}

void AutomapCSurf::setSelTrackHardSendMute(bool val)
{
	setBoolThreadSafe(&_selTrackHardSendMute, val);
}

bool AutomapCSurf::getSelTrackHardSendMute()
{
	return getBoolThreadSafe(&_selTrackHardSendMute);
}

void AutomapCSurf::setMidiViewMode(int view)
{
	setIntThreadSafe(&_midiViewMode, view);
}

int AutomapCSurf::getMidiViewMode()
{
	return getIntThreadSafe(&_midiViewMode);
}

void AutomapCSurf::setDefaultEnvShape(int shape)
{
	setIntThreadSafe(&_defaultEnvShape, shape);
}

int AutomapCSurf::getDefaultEnvShape()
{
	return getIntThreadSafe(&_defaultEnvShape);
}

void AutomapCSurf::setSelectedEnvShape(int shape)
{
	setIntThreadSafe(&_selectedEnvShape, shape);
}

int AutomapCSurf::getSelectedEnvShape()
{
	return getIntThreadSafe(&_selectedEnvShape);
}

void AutomapCSurf::setMidiGrid(int grid)
{
	setIntThreadSafe(&_midiGrid, grid);
}

int AutomapCSurf::getMidiGrid()
{
	return getIntThreadSafe(&_midiGrid);
}

void AutomapCSurf::setMainGrid(int grid)
{
	setIntThreadSafe(&_mainGrid, grid);
}

int AutomapCSurf::getMainGrid()
{
	return getIntThreadSafe(&_mainGrid);
}
