/*
** Novation Automap Control Surface for Cockos Reaper
** Copyright (C) 2009 Padre P.C. (Atrabilis Sunrise / P.C.M. / info@p-c-m.org)
** License: LGPL.
**
** This software is provided 'as-is', without any express or implied
** warranty.  In no event will the authors be held liable for any damages
** arising from the use of this software.
**
** Permission is granted to anyone to use this software for any purpose,
** including commercial applications, and to alter it and redistribute it
** freely, subject to the following restrictions:
**
** 1. The origin of this software must not be misrepresented; you must not
**    claim that you wrote the original software. If you use this software
**    in a product, an acknowledgment in the product documentation would be
**    appreciated but is not required.
** 2. Altered source versions must be plainly marked as such, and must not be
**    misrepresented as being the original software.
** 3. This notice may not be removed or altered from any source distribution.
**
** Notes: the C++ interfaces used require MSVC on win32, or at least the MSVC-compatible C++ ABI. Sorry, mingw users :(
**
*/

#include "AutomapDevice.h"
#include "AutomapCSurf.h"

AutomapController::AutomapController()
{
}

AutomapController::AutomapController(ReaperControlType type, int trackIndex, Type ctrlType)
 : _type(type), _trackIndex(trackIndex), _ctrlType(ctrlType)
{
	_paramIds.clear();
	_slots.clear();
	_value.clear();

	strcpy(this->name, "");

	switch(_ctrlType)
	{
		//! \bug Novation: 'a step size of zero creates a momentary button': can't make it work, I only get Normal/toggle
		case eBUTTON:
		case eMOMENTARY:
			maxInteger = 1;
			minInteger = 0;
			stepInteger = 1;
		break;
		case ePOT:
		case eENDLESS:
		default:
			//! \todo Check quantization step
			maxInteger = 100;
			minInteger = 0;
			stepInteger = 1;
		break;
	}
}

AutomapController::~AutomapController()
{
	_paramIds.clear();
	_slots.clear();
	_value.clear();
}

int AutomapController::getTrackIndex()
{
	return _trackIndex;
}

ReaperControlType AutomapController::getType()
{
	return _type;
}

void AutomapController::refreshClient(AutomapReaperClient *client, bool updateInfo)
{
	if(client == NULL)
		return;

	Automap::Connection* connection = client->GetConnection();
	if(connection == NULL)
		return;

	for(set<int>::iterator it = _paramIds.begin(); it != _paramIds.end(); it++)
	{
		if(updateInfo)
			connection->ParamInfoChanged(*it);
		else
			connection->SendEvent(*it, false);
	}
}

bool AutomapController::isEndless()
{
	bool isBadPtr = IsBadReadPtr(&_ctrlType, sizeof(Type)) == TRUE;
	if(isBadPtr)
		return false;
	return (_ctrlType == eENDLESS) || (_ctrlType == eMOMENTARY);
}

int AutomapController::getSlot(int paramId)
{
	return _slots[paramId];
}

void AutomapController::setSlot(int paramId, int slot)
{
	_slots[paramId] = slot;
}

float AutomapController::getValue(int slot)
{
	return _value[slot];
}

void AutomapController::setValue(float value, int slot)
{
	_value[slot] = value;
}

AutomapDevice::AutomapDevice()
{
}

AutomapDevice::AutomapDevice(bool hasGlobalPages)
: _hasGlobalPages(hasGlobalPages), _nbFxParamSlots(0), _nbTcpFxParamSlots(0), _nbTrackSendSlots(0), _nbTrackHardSendSlots(0)
, _currentPresetId(0)
//, _nbAssignableFnSlots(0)
{
	clearControllers();
	_nbPresets = 0;
	_nbParamsPerPage = 0;
	_nbTracksPerPage = 0;
	_nbGlobalPages = 0;
}

AutomapDevice::~AutomapDevice()
{
	clearControllers();
	_presetNames.clear();
}

AutomapController* AutomapDevice::addController(ReaperControlType type, int trackIndex, AutomapController::Type ctrlType, int paramId, int slot, float value)
{
	pair<ReaperControlType,int> key = make_pair(type, trackIndex);

	AutomapController* controller = _ctrlTypeIndexed[key];

	if(controller == NULL)
	{
		controller = new AutomapController(type, trackIndex, ctrlType);
		_ctrlTypeIndexed[key] = controller;
		_controllers.insert(controller);
	}

	controller->_paramIds.insert(paramId);
	controller->_slots[paramId] = slot;
	controller->_value[slot] = value;

	_ctrlParamIdIndexed[paramId] = controller;
	return controller;
}

void AutomapDevice::removeController(ReaperControlType type, int trackIndex, int paramId)
{
	pair<ReaperControlType,int> key = make_pair(type, trackIndex);
	AutomapController* controller = _ctrlTypeIndexed[key];

	if(controller)
	{
		if(paramId == -1)
		{
			for(set<int>::iterator it = controller->_paramIds.begin(); it != controller->_paramIds.end(); it++)
			_ctrlParamIdIndexed.erase(*it);
			controller->_paramIds.clear();
			controller->_slots.clear();
		}
		else
		{
			_ctrlParamIdIndexed.erase(paramId);
			controller->_paramIds.erase(paramId);
			controller->_slots.erase(paramId);
		}

		if(controller->_paramIds.empty())
		{
			_controllers.erase(controller);
			_ctrlTypeIndexed.erase(key);
			delete controller;
			controller = NULL;
		}
	}
}

AutomapController* AutomapDevice::getController(ReaperControlType type, int trackIndex)
{
	pair<ReaperControlType,int> key = make_pair(type, trackIndex);
	return _ctrlTypeIndexed[key];
}

AutomapController* AutomapDevice::getController(int paramId)
{
	return _ctrlParamIdIndexed[paramId];
}

void AutomapDevice::clearControllers()
{
	_ctrlTypeIndexed.clear();
	_ctrlParamIdIndexed.clear();
	for(set<AutomapController*>::iterator it = _controllers.begin(); it != _controllers.end(); it++)
	{
		if(*it != NULL)
		{
			delete *it;
			*it = NULL;
		}
	}
	_controllers.clear();
}

void AutomapDevice::initControllers(int nbTracks, int preset)
{
	if(preset != -1)
		_currentPresetId = preset;
//if(_currentPresetId >= _nbPresets)
//	_currentPresetId = _nbPresets - 1;

	clearControllers();
	initGlobalControllers();
	initTrackControllers(nbTracks);
}

int AutomapDevice::getNbPresets()
{
	return _nbPresets;
}

int AutomapDevice::getCurrentPreset(char name[AUTOMAP_NAME_LENGTH+1])
{
	if(_presetNames[_currentPresetId])
		AutomapReaperClient::strcpyCrop(name, _presetNames[_currentPresetId]);

	return _currentPresetId;
}

int AutomapDevice::getNbParams()
{
	int nbParams = this->getNbParamsPerPage();
	if(!_ctrlParamIdIndexed.empty())
	{
		map<int, AutomapController*>::iterator it = _ctrlParamIdIndexed.end();
		it--;
		nbParams = it->first + 1;
	}

	return nbParams;
}

int AutomapDevice::getNbFxParamSlots()
{
	return _nbFxParamSlots;
}

int AutomapDevice::getNbTcpFxParamSlots()
{
	return _nbTcpFxParamSlots;
}

int AutomapDevice::getNbTrackSendSlots()
{
	return _nbTrackSendSlots;
}

int AutomapDevice::getNbTrackHardSendSlots()
{
	return _nbTrackHardSendSlots;
}

//int AutomapDevice::getNbAssignableFnSlots()
//{
//	return _nbAssignableFnSlots;
//}

void AutomapDevice::getVolValueStr(double value, char* valueStr)
{
	value = SLIDER2DB(1000.0*value);
	if(value<-71.0)
		AutomapReaperClient::sprintfCrop(valueStr, "-inf dB");
	else
		AutomapReaperClient::sprintfCrop(valueStr, "%+2.1lfdB", value);
}

void AutomapDevice::getPanValueStr(double value, char* valueStr)
{
	value = 200.0*value - 100.0;

	if(fabs(value) < 0.08)
		value=0.0;

	if(value == 0.0)
		AutomapReaperClient::sprintfCrop(valueStr, "<C>");
	else if(value < 0.0)
		AutomapReaperClient::sprintfCrop(valueStr, "%02.0lf%%L", fabs(value));
	else
		AutomapReaperClient::sprintfCrop(valueStr, "%02.0lf%%R", fabs(value));
}

void AutomapDevice::getButtonValueStr(double value, char* valueStr)
{
	if(value<0.5)
		AutomapReaperClient::sprintfCrop(valueStr, "Off");
	else
		AutomapReaperClient::sprintfCrop(valueStr, "On");
}

void AutomapDevice::getTrackSelectValueStr(double value, char* valueStr)
{
	if(value<0.5)
		AutomapReaperClient::sprintfCrop(valueStr, " ");
	else
		AutomapReaperClient::sprintfCrop(valueStr, "^");
}

void AutomapDevice::getSubmenuValueStr(double value, char* valueStr)
{
	if(value<0.5)
		AutomapReaperClient::sprintfCrop(valueStr, " ");
	else
		AutomapReaperClient::sprintfCrop(valueStr, "X");
}

void AutomapDevice::getMcpModeValueStr(double value, char* valueStr)
{
	if(value<0.5)
		AutomapReaperClient::sprintfCrop(valueStr, "Track");
	else
		AutomapReaperClient::sprintfCrop(valueStr, "Mixer");
}

void AutomapDevice::getSendModeValueStr(int mode, char* valueStr)
{
	switch(mode)
	{
		case 0: AutomapReaperClient::strcpyCrop(valueStr, "Post-Fader");	break;
		case 1: AutomapReaperClient::strcpyCrop(valueStr, "Pre-FX");		break;
		case 2: AutomapReaperClient::strcpyCrop(valueStr, "???");			break;
		case 3: AutomapReaperClient::strcpyCrop(valueStr, "Pre-Fader");	break;
		default: AutomapReaperClient::strcpyCrop(valueStr, "???");			break;
	}
}

void AutomapDevice::getMidiViewModeValueStr(int view, char* valueStr)
{
	switch(view)
	{
		case 0	: AutomapReaperClient::strcpyCrop(valueStr, "Piano Roll");		break;
		case 1	: AutomapReaperClient::strcpyCrop(valueStr, "Drum Map");		break;
		case 2	: AutomapReaperClient::strcpyCrop(valueStr, "Event List");		break;
		default	: AutomapReaperClient::strcpyCrop(valueStr, "???");			break;
	}
}

int AutomapDevice::getMidiViewModeCommand(int view)
{
	switch(view)
	{
		case 0	: return ID_MIDI_VIEW_PIANOROLL;		break;
		case 1	: return ID_MIDI_VIEW_DRUMMAP;			break;
		case 2	: return ID_MIDI_VIEW_EVENTLIST;		break;
		default	: return -1;							break;
	}
}

void AutomapDevice::getGridValueStr(GridDivision grid, char* valueStr)
{
	switch(grid)
	{
		case eGRID_4		: AutomapReaperClient::strcpyCrop(valueStr, "4");		break;
		case eGRID_2		: AutomapReaperClient::strcpyCrop(valueStr, "2");		break;
		case eGRID_1		: AutomapReaperClient::strcpyCrop(valueStr, "1");		break;
		case eGRID_1_2		: AutomapReaperClient::strcpyCrop(valueStr, "1/2");		break;
		case eGRID_1_3		: AutomapReaperClient::strcpyCrop(valueStr, "1/3");		break;
		case eGRID_1_4		: AutomapReaperClient::strcpyCrop(valueStr, "1/4");		break;
		case eGRID_1_4T		: AutomapReaperClient::strcpyCrop(valueStr, "1/4T");	break;
		case eGRID_1_8		: AutomapReaperClient::strcpyCrop(valueStr, "1/8");		break;
		case eGRID_1_8T		: AutomapReaperClient::strcpyCrop(valueStr, "1/8T");	break;
		case eGRID_1_16		: AutomapReaperClient::strcpyCrop(valueStr, "1/16");	break;
		case eGRID_1_16T	: AutomapReaperClient::strcpyCrop(valueStr, "1/16T");	break;
		case eGRID_1_32		: AutomapReaperClient::strcpyCrop(valueStr, "1/32");	break;
		case eGRID_1_32T	: AutomapReaperClient::strcpyCrop(valueStr, "1/32T");	break;
		case eGRID_1_64		: AutomapReaperClient::strcpyCrop(valueStr, "1/64");	break;
		case eGRID_1_64T	: AutomapReaperClient::strcpyCrop(valueStr, "1/64T");	break;
		case eGRID_1_128	: AutomapReaperClient::strcpyCrop(valueStr, "1/128");	break;
		case eGRID_1_128T	: AutomapReaperClient::strcpyCrop(valueStr, "1/128T");	break;
		default				: AutomapReaperClient::strcpyCrop(valueStr, "???");		break;
	}
}

void AutomapDevice::getCcLaneValueStr(int lane, char* valueStr)
{
	switch(lane)
	{
		case 0		: AutomapReaperClient::strcpyCrop(valueStr, "Velocity");		break;
		case 1		: AutomapReaperClient::strcpyCrop(valueStr, "Pitch");			break;
		case 2		: AutomapReaperClient::strcpyCrop(valueStr, "Program");			break;
		case 3		: AutomapReaperClient::strcpyCrop(valueStr, "ChanPres");		break;
		case 4		: AutomapReaperClient::strcpyCrop(valueStr, "Bnk/Prog");		break;
		case 5		: AutomapReaperClient::strcpyCrop(valueStr, "TextEvt");			break;
		case 6		: AutomapReaperClient::strcpyCrop(valueStr, "Sysex");			break;
		default		:
			lane -= 7;
			if(lane<0)
				lane = 0;
			if(lane>119)
				lane = 119;
			AutomapReaperClient::sprintfCrop(valueStr, "CC#%03d", lane);
		break;
	}
}

void AutomapDevice::getAutomationModeValueStr(int mode, char* valueStr)
{
	switch(mode)
	{
		case 0	: AutomapReaperClient::strcpyCrop(valueStr, "Trim/Read");		break;
		case 1	: AutomapReaperClient::strcpyCrop(valueStr, "Read");			break;
		case 2	: AutomapReaperClient::strcpyCrop(valueStr, "Touch");			break;
		case 3	: AutomapReaperClient::strcpyCrop(valueStr, "Write");			break;
		case 4	: AutomapReaperClient::strcpyCrop(valueStr, "Latch");			break;
		default	: AutomapReaperClient::strcpyCrop(valueStr, "???");				break;
	}
}

void AutomapDevice::getEnvShapeValueStr(int shape, char* valueStr)
{
	switch(shape)
	{
		case 0	: AutomapReaperClient::strcpyCrop(valueStr, "Linear");			break;
		case 1	: AutomapReaperClient::strcpyCrop(valueStr, "Square");			break;
		case 2	: AutomapReaperClient::strcpyCrop(valueStr, "Slow");			break;
		case 3	: AutomapReaperClient::strcpyCrop(valueStr, "FastStart");		break;
		case 4	: AutomapReaperClient::strcpyCrop(valueStr, "FastEnd");			break;
		case 5	: AutomapReaperClient::strcpyCrop(valueStr, "Bezier");			break;
		default	: AutomapReaperClient::strcpyCrop(valueStr, "???");				break;
	}
}

int AutomapDevice::getDefaultEnvShapeCommand(int shape)
{
	switch(shape)
	{
		case 0	: return ID_MAIN_ENVELOPE_SET_DEFAULT_SHAPE_LINEAR;			break;
		case 1	: return ID_MAIN_ENVELOPE_SET_DEFAULT_SHAPE_SQUARE;			break;
		case 2	: return ID_MAIN_ENVELOPE_SET_DEFAULT_SHAPE_SLOW;			break;
		case 3	: return ID_MAIN_ENVELOPE_SET_DEFAULT_SHAPE_FASTSTART;		break;
		case 4	: return ID_MAIN_ENVELOPE_SET_DEFAULT_SHAPE_FASTEND;		break;
		case 5	: return ID_MAIN_ENVELOPE_SET_DEFAULT_SHAPE_BEZIER;			break;
		default	: return -1;												break;
	}
}

int AutomapDevice::getSelectedEnvShapeCommand(int shape)
{
	switch(shape)
	{
		case 0	: return ID_MAIN_ENVELOPE_SET_SELECTED_SHAPE_LINEAR;		break;
		case 1	: return ID_MAIN_ENVELOPE_SET_SELECTED_SHAPE_SQUARE;		break;
		case 2	: return ID_MAIN_ENVELOPE_SET_SELECTED_SHAPE_SLOW;			break;
		case 3	: return ID_MAIN_ENVELOPE_SET_SELECTED_SHAPE_FASTEND;		break;
		case 4	: return ID_MAIN_ENVELOPE_SET_SELECTED_SHAPE_FASTSTART;		break;
		case 5	: return ID_MAIN_ENVELOPE_SET_SELECTED_SHAPE_BEZIER;		break;
		default	: return -1;												break;
	}
}

int AutomapDevice::getNbParamsPerPage()
{
	return _nbParamsPerPage;
}

int AutomapDevice::getNbTracksPerPage()
{
	return _nbTracksPerPage;
}

int AutomapDevice::getNbGlobalPages()
{
	return _nbGlobalPages;
}

AutomapRemoteSl::AutomapRemoteSl(bool hasGlobalPages)
: AutomapDevice(hasGlobalPages)
{
	_nbParamsPerPage = AUTOMAP_REMOTESL_NBPARAMS_PER_PAGE;
	_nbTracksPerPage = AUTOMAP_REMOTESL_NBTRACKS_PER_PAGE;
	_nbGlobalPages = AUTOMAP_REMOTESL_NBGLOBALPAGES;
	initGlobalControllers();
}

AutomapRemoteSl::~AutomapRemoteSl()
{
}

AutomapRemoteSl::ParamType AutomapRemoteSl::getParamType(int paramId)
{
	int index = paramId%AUTOMAP_REMOTESL_NBPARAMS_PER_PAGE;
	return (ParamType)(index/8);
}

void AutomapRemoteSl::initGlobalControllers()
{
	_nbFxParamSlots = 8;
	_nbTcpFxParamSlots = 0;
	_nbTrackSendSlots = 1;
	_nbTrackHardSendSlots  = 1;
	//_nbAssignableFnSlots = 8;

	if(_hasGlobalPages)
	{
		addController(eCTRL_HZOOM_WHEEL,						-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER));
		addController(eCTRL_FFWRWD_WHEEL,						-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+1);
		//addController(eCTRL_PREVNEXTMARKER_WHEEL,				-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+2);
		addController(eCTRL_VSCROLL_WHEEL,						-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+2);
		addController(eCTRL_TRACKSELECT_WHEEL,					-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+3);
		addController(eCTRL_CURSEL_FXSELECT_WHEEL,				-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+4);
		addController(eCTRL_CURSEL_FXPARAMSELECT_WHEEL,			-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+5);
		addController(eCTRL_CURSEL_TRACKSENDSELECT_WHEEL,		-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+6);
		addController(eCTRL_CURSEL_TRACKHARDSENDSELECT_WHEEL,	-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+7);

		for(int slot=0; slot<8; slot++)
			addController(eCTRL_FXPARAM,			-1, AutomapController::ePOT,		8*(int)(ePARAM_SLIDER)+slot, slot);

		addController(eCTRL_MARKER_PREV,			-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1));
		addController(eCTRL_MARKER_NEXT,			-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+1);
		addController(eCTRL_GOSTART,				-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+2);
		addController(eCTRL_GOEND,					-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+3);
		addController(eCTRL_LOOP_SETSTART,			-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+4);
		addController(eCTRL_LOOP_SETEND,			-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+5);
		addController(eCTRL_INSERT_MARKER,			-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+6);
		//addController(eCTRL_MCPMODE,				-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+7);
		//addController(eCTRL_INSERT_MARKERRGN,		-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+7);
		addController(eCTRL_CLEARRECARM,			-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+7);

		//addController(eCTRL_RWD,					-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1));
		//addController(eCTRL_FFW,					-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+1);
		//addController(eCTRL_GOSTART,				-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+2);
		//addController(eCTRL_GOEND,				-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+3);
		//addController(eCTRL_STOP,					-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+4);
		//addController(eCTRL_PLAY,					-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+5);
		//addController(eCTRL_LOOP,					-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+6);
		//addController(eCTRL_RECORD,				-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_LEFT1)+7);

		for(int slot=0; slot<8; slot++)
			addController(eCTRL_ASSIGNABLEFN,		-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON_RIGHT)+slot, slot);

		addController(eCTRL_VOL,					0, AutomapController::ePOT,		8*(int)(ePARAM_POT));
		addController(eCTRL_PAN,					0, AutomapController::ePOT,		8*(int)(ePARAM_POT)+1);
		addController(eCTRL_CURSEL_VOL,				-1, AutomapController::ePOT,	8*(int)(ePARAM_POT)+2);
		addController(eCTRL_CURSEL_PAN,				-1, AutomapController::ePOT,	8*(int)(ePARAM_POT)+3);
		addController(eCTRL_TRACKSEND_VOL,			-1, AutomapController::ePOT,	8*(int)(ePARAM_POT)+4);
		addController(eCTRL_TRACKSEND_PAN,			-1, AutomapController::ePOT,	8*(int)(ePARAM_POT)+5);
		addController(eCTRL_TRACKHARDSEND_VOL,		-1, AutomapController::ePOT,	8*(int)(ePARAM_POT)+6);
		addController(eCTRL_TRACKHARDSEND_PAN,		-1, AutomapController::ePOT,	8*(int)(ePARAM_POT)+7);

		//addController(eCTRL_MUTE,					0, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2));
		//addController(eCTRL_SOLO,					0, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2)+1);
		//addController(eCTRL_CLEARRECARM,			-1, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2));
		addController(eCTRL_CURSEL_SOLO,			-1, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2));
		addController(eCTRL_CURSEL_MUTE,			-1, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2)+1);
		addController(eCTRL_CURSEL_FXBYPASS,		-1, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2)+2);
		addController(eCTRL_TRACKSEND_MUTE,			-1, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2)+4);
		addController(eCTRL_TRACKSEND_MODE,			-1, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2)+5);
		addController(eCTRL_TRACKHARDSEND_MUTE,		-1, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2)+6);
		addController(eCTRL_TRACKHARDSEND_MODE,		-1, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2)+7);
	}
}

void AutomapRemoteSl::initTrackControllers(int nbTracks)
{
	int nbPages = nbTracks/AUTOMAP_REMOTESL_NBTRACKS_PER_PAGE + 1;

	for(int page=0; page<nbPages; page++)
	{
		int pageOffset = page*AUTOMAP_REMOTESL_NBPARAMS_PER_PAGE;
		if(_hasGlobalPages)
			pageOffset += AUTOMAP_REMOTESL_NBGLOBALPAGES*AUTOMAP_REMOTESL_NBPARAMS_PER_PAGE;
		for(int trackOffset=0; trackOffset<8; trackOffset++)
		{
			int track = page*AUTOMAP_REMOTESL_NBTRACKS_PER_PAGE + trackOffset + 1;
			if(track<=nbTracks)
			{
				if(trackOffset==0)	// Global controllers
				{
					for(int slot=0; slot<8; slot++)
						addController(eCTRL_ASSIGNABLEFNPOT,	-1, AutomapController::ePOT, 8*(int)(ePARAM_ENCODER)+trackOffset+pageOffset+slot, slot);
				}

				addController(eCTRL_VOL,			track, AutomapController::ePOT,		8*(int)(ePARAM_SLIDER)+trackOffset+pageOffset);
				addController(eCTRL_PAN,			track, AutomapController::ePOT,		8*(int)(ePARAM_POT)+trackOffset+pageOffset);
				addController(eCTRL_SOLO,			track, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT1)+trackOffset+pageOffset);
				addController(eCTRL_MUTE,			track, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_RIGHT)+trackOffset+pageOffset);
				addController(eCTRL_ARM,			track, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON_LEFT2)+trackOffset+pageOffset);
			}
		}
	}
}

AutomapiPhone::AutomapiPhone(bool hasGlobalPages)
: AutomapDevice(hasGlobalPages)
{
	_nbParamsPerPage = AUTOMAP_IPHONE_NBPARAMS_PER_PAGE;
	_nbTracksPerPage = AUTOMAP_IPHONE_NBTRACKS_PER_PAGE;
	_nbGlobalPages = AUTOMAP_IPHONE_NBGLOBALPAGES;
	initGlobalControllers();
}

AutomapiPhone::~AutomapiPhone()
{
}

AutomapiPhone::ParamType AutomapiPhone::getParamType(int paramId)
{
	int index = paramId%AUTOMAP_IPHONE_NBPARAMS_PER_PAGE;
	if(index<2)
		return ePARAM_SLIDER;
	else
		return ePARAM_BUTTON;
}

void AutomapiPhone::initGlobalControllers()
{
	_nbFxParamSlots = 0;
	_nbTcpFxParamSlots = 0;
	_nbTrackSendSlots = 0;
	_nbTrackHardSendSlots  = 0;
	//_nbAssignableFnSlots = 4;

	if(_hasGlobalPages)
	{
		addController(eCTRL_VOL,				0, AutomapController::ePOT,	0);
		addController(eCTRL_PAN,				0, AutomapController::ePOT,	1);
		addController(eCTRL_MUTE,				0, AutomapController::eBUTTON,	2);
		addController(eCTRL_SOLO,				0, AutomapController::eBUTTON,	3);
		addController(eCTRL_RWD,				-1, AutomapController::eMOMENTARY,	4);
		addController(eCTRL_FFW,				-1, AutomapController::eMOMENTARY,	5);
		addController(eCTRL_STOP,				-1, AutomapController::eMOMENTARY,	6);
		addController(eCTRL_PLAY,				-1, AutomapController::eMOMENTARY,	7);
		addController(eCTRL_LOOP,				-1, AutomapController::eMOMENTARY,	8);
		addController(eCTRL_RECORD,				-1, AutomapController::eMOMENTARY,	9);

		int pageOffset = AUTOMAP_IPHONE_NBPARAMS_PER_PAGE;

		//addController(eCTRL_ZOOM_WHEEL,	-1, AutomapController::eENDLESS,	pageOffset+0);
		//addController(eCTRL_MASTERPAN,	-1, AutomapController::ePOT,		pageOffset+1);

		for(int slot=0; slot<8; slot++)
			addController(eCTRL_ASSIGNABLEFN,						-1, AutomapController::eMOMENTARY, pageOffset+2+slot, slot);
	}
}

void AutomapiPhone::initTrackControllers(int nbTracks)
{
	int nbPages = nbTracks/AUTOMAP_IPHONE_NBTRACKS_PER_PAGE + 1;

	for(int page=0; page<nbPages; page++)
	{
		int pageOffset = page*AUTOMAP_IPHONE_NBPARAMS_PER_PAGE;
		if(_hasGlobalPages)
			pageOffset += AUTOMAP_IPHONE_NBGLOBALPAGES*AUTOMAP_IPHONE_NBPARAMS_PER_PAGE;

		int track = page*AUTOMAP_IPHONE_NBTRACKS_PER_PAGE + 1;
		if(track<=nbTracks)
		{
			addController(eCTRL_VOL,			track, AutomapController::ePOT,		pageOffset);
			addController(eCTRL_PAN,			track, AutomapController::ePOT,		pageOffset+1);
			addController(eCTRL_SOLO,			track, AutomapController::eBUTTON,	pageOffset+2);
			addController(eCTRL_MUTE,			track, AutomapController::eBUTTON,	pageOffset+3);
			addController(eCTRL_ARM,			track, AutomapController::eBUTTON,	pageOffset+4);

			addController(eCTRL_PLAY,			-1, AutomapController::eMOMENTARY,	pageOffset+6);
			addController(eCTRL_STOP,			-1, AutomapController::eMOMENTARY,	pageOffset+7);

			for(int slot=0; slot<2; slot++)
				addController(eCTRL_ASSIGNABLEFN,	-1, AutomapController::eMOMENTARY, pageOffset+8+slot, slot);
		}
	}
}

AutomapNocturn::AutomapNocturn(bool hasGlobalPages)
: AutomapDevice(hasGlobalPages)
{
	_nbParamsPerPage = AUTOMAP_NOCTURN_NBPARAMS_PER_PAGE;
	_nbTracksPerPage = AUTOMAP_NOCTURN_NBTRACKS_PER_PAGE;
	_nbGlobalPages = AUTOMAP_NOCTURN_NBGLOBALPAGES;
	initGlobalControllers();
}

AutomapNocturn::~AutomapNocturn()
{
}

AutomapNocturn::ParamType AutomapNocturn::getParamType(int paramId)
{
	int index = paramId%AUTOMAP_NOCTURN_NBPARAMS_PER_PAGE;
	return (ParamType)(index/8);
}

void AutomapNocturn::initGlobalControllers()
{
	_nbFxParamSlots = 0;
	_nbTcpFxParamSlots = 0;
	_nbTrackSendSlots = 0;
	_nbTrackHardSendSlots  = 0;
	//_nbAssignableFnSlots = 4;

	if(_hasGlobalPages)
	{
		int pageOffset = 0;

		addController(eCTRL_VOL,					0, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER));
		addController(eCTRL_PAN,					0, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+1);
		addController(eCTRL_HZOOM_WHEEL,			-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+2);
		addController(eCTRL_FFWRWD_WHEEL,			-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+3);
		addController(eCTRL_PREVNEXTMARKER_WHEEL,	-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+4);
		addController(eCTRL_TRACKSELECT_WHEEL,		-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+5);
		//addController(eCTRL_FXSELECT_WHEEL,		-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+6);
		//addController(eCTRL_FXPARAM1,				-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+7);

		addController(eCTRL_LOOP_SETSTART,			-1, AutomapController::eMOMENTARY,		8*(int)(ePARAM_BUTTON));
		addController(eCTRL_LOOP_SETEND,			-1, AutomapController::eMOMENTARY,		8*(int)(ePARAM_BUTTON)+1);
		addController(eCTRL_INSERT_MARKERRGN,		-1, AutomapController::eMOMENTARY,		8*(int)(ePARAM_BUTTON)+2);
		addController(eCTRL_INSERT_MARKER,			-1, AutomapController::eMOMENTARY,		8*(int)(ePARAM_BUTTON)+3);

		for(int slot=0; slot<4; slot++)
			addController(eCTRL_ASSIGNABLEFN,		-1, AutomapController::eMOMENTARY, 8*(int)(ePARAM_BUTTON)+4+slot, slot);

		pageOffset = AUTOMAP_NOCTURN_NBPARAMS_PER_PAGE;

		addController(eCTRL_CURSEL_SOLO,			-1, AutomapController::eBUTTON,		8*(int)(ePARAM_BUTTON)+pageOffset);
		addController(eCTRL_CURSEL_MUTE,			-1, AutomapController::eBUTTON,		8*(int)(ePARAM_BUTTON)+pageOffset+1);
		addController(eCTRL_CURSEL_ARM,				-1, AutomapController::eBUTTON,		8*(int)(ePARAM_BUTTON)+pageOffset+2);
		addController(eCTRL_CURSEL_FXBYPASS,		-1, AutomapController::eBUTTON,		8*(int)(ePARAM_BUTTON)+pageOffset+3);
		addController(eCTRL_CURSEL_PHASE,			-1, AutomapController::eBUTTON,		8*(int)(ePARAM_BUTTON)+pageOffset+4);
		addController(eCTRL_CURSEL_VIEW_IO,			-1, AutomapController::eMOMENTARY,	8*(int)(ePARAM_BUTTON)+pageOffset+5);
		addController(eCTRL_CURSEL_VIEW_ENVELOPES,	-1, AutomapController::eMOMENTARY,	8*(int)(ePARAM_BUTTON)+pageOffset+6);
		addController(eCTRL_CURSEL_VIEW_FX,			-1, AutomapController::eMOMENTARY,	8*(int)(ePARAM_BUTTON)+pageOffset+7);

		addController(eCTRL_CURSEL_VOL,					-1, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+pageOffset);
		addController(eCTRL_CURSEL_PAN,					-1, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+pageOffset+1);
		addController(eCTRL_SOLO_INFRONT_ADJUST_POT,	-1, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+pageOffset+2);
		addController(eCTRL_HSCROLL_POT,				-1, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+pageOffset+3);
		addController(eCTRL_VSCROLL_POT,				-1, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+pageOffset+4);
		addController(eCTRL_HZOOM_POT,					-1, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+pageOffset+5);
		addController(eCTRL_VZOOM_POT,					-1, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+pageOffset+6);
		addController(eCTRL_TRACKSELECT_WHEEL,			-1, AutomapController::eENDLESS,	8*(int)(ePARAM_ENCODER)+pageOffset+7);
	}
}

void AutomapNocturn::initTrackControllers(int nbTracks)
{
	int nbPages = nbTracks/AUTOMAP_NOCTURN_NBTRACKS_PER_PAGE + 1;

	for(int page=0; page<nbPages; page++)
	{
		int pageOffset = page*AUTOMAP_NOCTURN_NBPARAMS_PER_PAGE;
		if(_hasGlobalPages)
			pageOffset += AUTOMAP_NOCTURN_NBGLOBALPAGES*AUTOMAP_NOCTURN_NBPARAMS_PER_PAGE;

		for(int trackOffset=0; trackOffset<AUTOMAP_NOCTURN_NBTRACKS_PER_PAGE; trackOffset++)
		{
			int track = page*AUTOMAP_NOCTURN_NBTRACKS_PER_PAGE + trackOffset + 1;
			if(track<=nbTracks)
			{
				addController(eCTRL_VOL,			track, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+2*trackOffset+pageOffset);
				addController(eCTRL_PAN,			track, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER)+1+2*trackOffset+pageOffset);
				addController(eCTRL_SOLO,			track, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON)+2*trackOffset+pageOffset);
				addController(eCTRL_MUTE,			track, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON)+1+2*trackOffset+pageOffset);
			}
		}
	}
}

AutomapRemoteSlCompact::AutomapRemoteSlCompact(bool hasGlobalPages)
: AutomapRemoteSl(hasGlobalPages)
{
}

AutomapRemoteSlCompact::~AutomapRemoteSlCompact()
{
}

AutomapRemoteSlCompact::ParamType AutomapRemoteSlCompact::getParamType(int paramId)
{
	int index = paramId%AUTOMAP_REMOTESL_NBPARAMS_PER_PAGE;
	return (ParamType)(index/8);
}

void AutomapRemoteSlCompact::initTrackControllers(int nbTracks)
{
	int nbPages = nbTracks/AUTOMAP_REMOTESL_NBTRACKS_PER_PAGE + 1;

	for(int page=0; page<nbPages; page++)
	{
		int pageOffset = page*AUTOMAP_REMOTESL_NBPARAMS_PER_PAGE;
		if(_hasGlobalPages)
			pageOffset += AUTOMAP_REMOTESL_NBGLOBALPAGES*AUTOMAP_REMOTESL_NBPARAMS_PER_PAGE;
		for(int trackOffset=0; trackOffset<8; trackOffset++)
		{
			int track = page*AUTOMAP_REMOTESL_NBTRACKS_PER_PAGE + trackOffset + 1;
			if(track<=nbTracks)
			{
				addController(eCTRL_VOL,			track, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER1)+trackOffset+pageOffset);
				addController(eCTRL_PAN,			track, AutomapController::ePOT,		8*(int)(ePARAM_ENCODER2)+trackOffset+pageOffset);
				addController(eCTRL_SOLO,			track, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON1)+trackOffset+pageOffset);
				addController(eCTRL_MUTE,			track, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON2)+trackOffset+pageOffset);
				addController(eCTRL_ARM,			track, AutomapController::eBUTTON,	8*(int)(ePARAM_BUTTON3)+trackOffset+pageOffset);
			}
		}
	}
}

