/*
** Novation Automap Control Surface for Cockos Reaper
** Copyright (C) 2009 Padre P.C. (Atrabilis Sunrise / P.C.M. / info@p-c-m.org)
** License: LGPL.
**
** This software is provided 'as-is', without any express or implied
** warranty.  In no event will the authors be held liable for any damages
** arising from the use of this software.
**
** Permission is granted to anyone to use this software for any purpose,
** including commercial applications, and to alter it and redistribute it
** freely, subject to the following restrictions:
**
** 1. The origin of this software must not be misrepresented; you must not
**    claim that you wrote the original software. If you use this software
**    in a product, an acknowledgment in the product documentation would be
**    appreciated but is not required.
** 2. Altered source versions must be plainly marked as such, and must not be
**    misrepresented as being the original software.
** 3. This notice may not be removed or altered from any source distribution.
**
** Notes: This is a small app to retrieve Automap parameter index information for proper mapping
**
*/

#include "stdio.h"
#include "windows.h"

#include "AutomapSDK/include/Automap.h"
using namespace Automap;


class Win32CriticalSection
{
	public:
		Win32CriticalSection(void);
		virtual ~Win32CriticalSection(void);

		virtual bool TryEnter(void);
		virtual bool Enter(int);
		virtual void Exit(void);
};

class ICriticalSection : public Win32CriticalSection
{
	public:
		ICriticalSection(void);
		virtual ~ICriticalSection(void);
};

class Win32Mutex
{
	public:
		Win32Mutex(char const *);
		virtual ~Win32Mutex(void);

		virtual bool Lock(int);
		virtual void Unlock(void);
};

class IMutex : public Win32Mutex
{
	public:
		IMutex(void);
		virtual ~IMutex(void);
};

ICriticalSection* __stdcall CreateICriticalSection(void);
IMutex * __stdcall CreateIMutex(char const *);
int __stdcall GetClientProcessID(void);
bool __stdcall IsProcessAlive(int);
char const * __stdcall GetServerExecutable(char *);


#define NUMPARAMS	90

//! \note Max ParamId is 2048
//! \brief Small client app to figure out ParamId mappings for a given device
class ParamIdClient : public Client
{
//	private:
public:
		Connection*	_connection;
		int _numParams;
bool changeMap;

	public:
		ParamIdClient();
		~ParamIdClient();

		void GetName(char *name) const
		{
			sprintf(name, "ParamIdClient");
		}

		void GetMfr(char *mfr) const
		{
			sprintf(mfr, "Padre P.C.");
		}

		void GetType(char *type) const
		{
			sprintf(type, "User");
		}

		void GetSaveName(char *name) const
		{
			sprintf(name, "ParamIdClient");
		}

		Automap::GUID GetGUID() const
		{
			return Automap::GUID("{78D444CB-0C00-4cdb-8B24-0EF2E57110B0}");
		}
		
		void GetInstanceName(char *name) const
		{
			sprintf(name, "ParamId");
		}

		void SaveInstanceName(const char *name)
		{
		}
		
		int GetNumParams() const
		{
			return _numParams;
		}

		Connection*	GetConnection()
		{
			return _connection;
		}
		
		bool GetParamInfo(int idx, ParamInfo& data);
		float GetParamValue(int idx, char valueTextOut[AUTOMAP_NAME_LENGTH+1]);
		void SetParamValue(int idx, float value);

		int GetNumPresets() const;
		int GetCurrentPreset(char name[AUTOMAP_NAME_LENGTH+1]) const;
		void SetCurrentPreset(int n);
};

ParamIdClient::ParamIdClient()
: _connection(NULL), _numParams(NUMPARAMS)
, changeMap(false)
{
	_connection = Connect(this);
	_connection->Initialize();		
	_connection->SetAutoFocus();
}

ParamIdClient::~ParamIdClient()
{
	Disconnect(_connection);
}

bool ParamIdClient::GetParamInfo(int idx, ParamInfo& data)
{
if(idx>=_numParams)
return false;
	sprintf(data.name, "p%d", idx);
	data.maxInteger = 1;
	data.minInteger = 0;
	data.stepInteger = 1;

	return true;
}

//float ParamIdClient::GetParamValue(int idx, char *valueTextOut)
float ParamIdClient::GetParamValue(int idx, char valueTextOut[AUTOMAP_NAME_LENGTH+1])
{
	return 0.5f;
}

void ParamIdClient::SetParamValue(int idx, float value)
{
changeMap = true;
}

int ParamIdClient::GetNumPresets() const
{
	return 0;
}

int ParamIdClient::GetCurrentPreset(char name[AUTOMAP_NAME_LENGTH+1]) const
{
	return -1;
}

void ParamIdClient::SetCurrentPreset(int n)
{
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HDC hdc;

	switch(uMsg)
	{
		case WM_PAINT:
			hdc = BeginPaint(hWnd, &ps);
			TextOut(hdc, 15, 15, (LPCWSTR)L"Automap Param Id tool is running.", 33);
			TextOut(hdc, 15, 45, (LPCWSTR)L"Please take screenshots of your Automap mappings.", 49);
			EndPaint(hWnd, &ps);
		break;

		case WM_CLOSE:
            DestroyWindow(hWnd);
        break;

        case WM_DESTROY:
            PostQuitMessage(0);
        break;

        default:
            return DefWindowProc(hWnd, uMsg, wParam, lParam);

	}

	return 0;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	ParamIdClient client;

    WNDCLASSEX wc;
    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         = 0;
    wc.lpfnWndProc   = WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = (LPCWSTR)L"myWindowClass";
    wc.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);

    if(!RegisterClassEx(&wc))
    {
		MessageBox(NULL, (LPCWSTR)L"Window Registration Failed!", (LPCWSTR)L"Error!", MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }

	HWND hwnd = CreateWindowEx(WS_EX_CLIENTEDGE, (LPCWSTR)L"myWindowClass", (LPCWSTR)L"Automap Param Id Tool", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, 400, 120, NULL, NULL, hInstance, NULL);

    ShowWindow(hwnd, nCmdShow);
    UpdateWindow(hwnd);

	MSG msg;
	while(GetMessage(&msg, NULL, 0, 0) > 0)
	{
		client.GetConnection()->Update();

Sleep(500);
if(client.changeMap)
{
client._numParams = 20;
client.GetConnection()->Initialize();


//ICriticalSection* sec = CreateICriticalSection();
//
//char serverExeName[1024];
//GetServerExecutable(serverExeName);
//IMutex* mutex = CreateIMutex(serverExeName);
//
//int processId = GetClientProcessID();
//if(IsProcessAlive(processId) && sec->Enter(processId))
//{
//client.GetConnection()->ParamCountChanged();
//sec->Exit();
//}


//if(mutex->Lock(processId))
//{
client.GetConnection()->ParamCountChanged();
//mutex->Unlock();
//}




//sec->TryEnter()
client.changeMap = false;
}

		Sleep(10);

		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	
	return (int)(msg.wParam);
}

